#!/usr/bin/env python
# -*- coding: utf8 -*-
"""
USAGE
    python management_pack.py --export [filename]
        The filename defaults to mp_dump.json

    python management_pack.py --import [--force] [filename]
        The filename defaults to mp_dump.json

        The import script is acting cowardly and refuses to overwrite any
        existing management pack, unless you pass the --force parameter

    python management_pack.py
        Start an interactive session to create one or more management packs

REQUIREMENT
    requests (yum install python-pip; sudo pip install requests)"""

import getpass
import json
import sys
try:
    import requests
except ImportError as e:
    print("ERROR\nIn this example we use a python library called requests which makes it easy to perform HTTP requests. requests is not installed together with python so you have to install it yourself. If you are using yum, yum install python-requests should be enough. Python savvy people might like to use pip instead: pip install requests, but the choice is up to you. We recommend you to read up on the subject if python is your preferred language.")
    sys.exit(1)

base_url = "https://your-server"
auth = ('user', 'password')
verify_ssl_cert = False
default_filename = 'mp_dump.json'

class ManagementPack(object):
    verify_ssl_cert = False

    def __init__(self, base_url, auth, verify_ssl_cert=None):
        self.base_url = base_url
        self.auth = auth
        if verify_ssl_cert:
            self.verify_ssl_cert = verify_ssl_cert


    def _get_all(self):
        """ Parses the complete list of management packs and then traverses
        all of the resources """
        listing = requests.get(self.base_url + '/api/config/management_pack', auth=self.auth, headers={'accept': 'application/json'}, verify=self.verify_ssl_cert)
        if listing.status_code >= 300:
            raise Exception("We got an invalid response (%d):\n%s" % (listing.status_code, listing.text))

        packs = []
        for management_pack in listing.json():
            pack = requests.get(management_pack['resource'], auth=self.auth, headers={'accept': 'application/json'}, verify=self.verify_ssl_cert)
            if pack.status_code >= 300:
                raise Exception("We got an invalid response (%d):\n%s" % (pack.status_code, pack.text))
            packs.append(pack.json())
        return packs

    def packs_to_json_file(self, filename):
        with open(filename, 'w') as out:
            out.write(json.dumps(self._get_all(), indent=4))

    def send_pack(self, pack, force=None):
        """ Communicates with the HTTP API,
        might raise exceptions """
        missing_props = set((u'description', u'hostgroup', u'icon', u'management_pack_name')) - set(pack)
        if missing_props:
            pack_name = ""
            if 'management_pack_name' in pack:
                pack_name = pack['management_pack_name']
            raise Exception("Invalid keys in JSON, not importing\n%s also needs %s" % (pack_name, ', '.join(missing_props)))
        uri = self.base_url + '/api/config/management_pack'
        push_method = requests.post
        if force:
            uri += '/' + pack['management_pack_name']
            push_method = requests.put
        result = push_method(uri, auth=self.auth, data=pack, headers={'accept': 'application/json'}, verify=self.verify_ssl_cert)
        if result.status_code >= 300:
            raise Exception("%d:\n%s" % (result.status_code, result.text))

    def json_file_to_api(self, filename, force=None):
        packs = json.loads(open(filename).read())
        if isinstance(packs, dict):
            packs = [packs]
        for pack in packs:
            self.send_pack(pack, force)

def interactive():
    """ Could be translated into something object type-agnostic when the
    API provides the properties table in a better format (i.e. discoverable help) """
    def raw_default(query, default, hide_input=None):
        """ Ask a question, accept empty answer but replace it with default """
        if hide_input:
            answer = getpass.getpass(query + "\n")
        else:
            answer = raw_input(query + "\n")
        if not answer:
            answer = default
        return answer
    def require_answer(query):
        """ Asks until answered """
        answer = ""
        while answer == "":
            answer = raw_input(query + "\n").strip()
        return answer
    def confirm(query, answer):
        """ Handles question? [y/N] """
        return answer.lower() == raw_input(query + "\n").lower()

    def custom_vars():
        if not confirm("Do you want to pass any custom variables to your hostgroup? [y/N]", 'y'):
            return {}
        variables = {}
        while True:
            try:
                variable = require_answer("What's the custom variable?").upper()
                if not variable.startswith('_'):
                    variable = '_' + variable
                variables[variable] = require_answer("How would you describe the custom variable?")
                print("These are the current custom variables:")
                for var in variables:
                    print("%s\n\t%s" % (var.upper(), variables[var]))
                if not confirm("Do you want to add any more custom variables to your hostgroup? [y/N]", 'y'):
                    break
            except KeyboardInterrupt:
                # Let the user get away with a bad patience :)
                break
        return variables

    try:
        url = raw_default("Where is your monitor server? [%s]" % (base_url), base_url)
        username = raw_default("Which username do you want to auth with? [%s]" % (auth[0]), auth[0])
        password = raw_default("And which password do you want to use?%s" % (" [saved in file]" if auth[1] else ""), auth[1], True)
    except KeyboardInterrupt:
        return
    model = ManagementPack(url, (username, password))
    while True:
        try:
            # @TODO validate against existing?
            monitor_object = {
                'management_pack_name': require_answer("What do you want to name the management pack?"),
                'description': require_answer("How would you describe the management pack?"),
                'hostgroup': require_answer("Which hostgroup do you want to use?"),
                'icon': require_answer("Which icon do you want to use?")
            }
            monitor_object.update(custom_vars())
            print("\nThis is what want to create:\n")
            for key in monitor_object:
                print("{0:20}: {1:20}".format(key, monitor_object[key]))
            if not confirm("Should I create it for you now? [Y/n]", 'n'):
                try:
                    force = confirm("If '%s' already exists, do you want to overwrite it? [y/N]" % (monitor_object['management_pack_name'],), 'y')
                    model.send_pack(monitor_object, force=force)
                    print("Management pack '%s' created!" % (monitor_object['management_pack_name'],))
                except Exception as e:
                    print(e)
            if 'y' != raw_input("Do you want to create another management pack? [y/N]").lower():
                break
        except KeyboardInterrupt:
            # Let the user get away with a bad patience :)
            break


if __name__ == "__main__":
    argv = sys.argv[1:]
    filename = default_filename
    management_pack = ManagementPack(base_url, auth, verify_ssl_cert)

    if '--export' in argv:
        argv.remove('--export')
        if argv:
            filename = argv[0]
        try:
            management_pack.packs_to_json_file(filename)
            print("Successfully exported to " + filename)
        except Exception as e:
            print(e)
            sys.exit(1)
    elif '--import' in argv:
        argv.remove('--import')
        force = False
        if '--force' in argv:
            argv.remove('--force')
            force = True
        if argv:
            filename = argv[0]
        try:
            management_pack.json_file_to_api(filename, force)
            print("Successfully imported " + filename)
        except Exception as e:
            print(e)
            sys.exit(1)
    elif '-h' in argv or '--help' in argv:
        print(__doc__)
    elif sys.__stdin__.isatty():
        interactive()
    else:
        print(__doc__)
        sys.exit(1)
