<?php

/**
 * @param $request Api_Request
 * @param $auth Op5Auth
 * @return Api_Response
 */
function generate_response(Api_Request $request, Op5Auth $auth) {
	$api = new Filter_Api($auth);
	return $api->handle_request($request);
}

class Filter_Api {
	// @TODO: POST to save filters
	// @TODO: DELETE to remove a saved filter
	// @TODO: PUT to replace a saved filter
	private $auth;
	private $limit;
	private $query;
	private $offset = array(); // @todo int?
	private $columns = array();
	private $sort = array();

	const DEFAULT_LIMIT = 50;

	/**
	 * @param $auth Op5Auth
	 */
	function __construct(Op5Auth $auth) {
		$this->auth = $auth;
	}

	protected function get_menu() {
		return array(
			array(
				'title' => 'Sections',
				'items' => array(
					'count' => array(
						'resource' => '/api/help/filter/count'
					),
					'query' => array(
						'resource' => '/api/help/filter/query'
					)
				)
			)
		);
	}

	/**
	 * Main entry point
	 *
	 * @todo AUTH?!
	 *
	 * @param $request Api_Request
	 * @returns Api_Response
	 * @throws Api_Error_Response
	 */
	function handle_request(Api_Request $request) {
		$path = $request->get_path();
		$method = array_shift($path);
		if($request->is_help()) {
			$response = array();
			$properties = array();
			switch($method) {
				case 'query':
					$description = file_get_contents(__DIR__.'/filter/query.html');
					$properties = array(
						array(
							'name' => 'query',
							'type' => 'string',
							'default' => '',
							'description' => 'Only find objects matching this query, for example "hosts that are down": <em>[hosts] state = 1</em>.',
							'required' => true
						),
						array(
							'name' => 'columns',
							'type' => 'string',
							'default' => '',
							'description' => "Comma separated list of columns which you want to include in the output. This makes sure you keep the response data to its minumum. For example: <em>columns=host.name,description</em> appended to a query for the service table, would only return the services' complete name.",
							'required' => false
						),
						array(
							'name' => 'limit',
							'type' => 'int',
							'default' => self::DEFAULT_LIMIT,
							'description' => 'Include at most <em>limit</em> objects in the result.',
							'required' => false
						),
						array(
							'name' => 'offset',
							'type' => 'int',
							'default' => '0',
							'description' => 'Skip the first <em>outset</em> results. Useful for pagination.',
							'required' => false
						),
						array(
							'name' => 'sort',
							'type' => 'string',
							'default' => '',
							'description' => 'Comma separated list of columns that dictates the objects order, for example <em>sort=name</em>; <em>sort=name,state</em> or <em>sort=name DESC,state ASC</em>. Note that you cannot sort on "lists", such as the "parents" field in the "hosts" table.',
							'required' => false
						),
					);
					break;
				case 'count':
					$description = file_get_contents(__DIR__.'/filter/count.html');
					$properties = array(
						array(
							'name' => 'query',
							'type' => 'string',
							'default' => '',
							'description' => 'Only find objects matching this query, for example "hosts that are down": <em>[hosts] state = 1</em>.',
							'required' => true
						)
					);
					break;
				case '':
					$description = file_get_contents(__DIR__.'/filter/filter.html');
					break;
				default:
					throw new Api_Error_Response('Help section not found', 404, "'$method' is not a valid filter API section");

			}
			return new Api_Response(
				array(
					'description' => $description,
					'body' => '',
					'menu' => $this->get_menu(),
					'properties' => $properties
				)
			);
		}

		if($method == '') {
			$body = array(
				'query' => array(
					'description' => "Ask Monitor what it knows about your system in this very moment. Filters are a made up query language which let's you optimize your interaction with Monitor.",
					'resource' => '/api/filter/query'
				),
				'count' => array(
					'description' => "Ask for the number of objects matching a specific query. If you are new to filters, you should probably check out the 'query' section first.",
					'resource' => '/api/filter/count'
				),
			);
			return new Api_Response($body);
		}

		if($method != 'count' && $method != 'query') {
			return new Api_Error_Response("Invalid filter method '$method', expecting count or query", 404, 'For example /api/filter/count');
		}

		if($errors = $this->set_parameters($request->get_params())) {
			return new Api_Error_Response("Invalid parameters", 400, $errors);
		}

		try {
			$set = ObjectPool_Model::get_by_query($this->query);
			// mayi_resource is, for example, 'monitor.monitoring.hosts'
			mayi_run_api($set->mayi_resource().':read.%s.status');
			$result = array();

			$pool = ObjectPool_Model::pool($set->get_table());
			if(!$pool->available_for('api')) {
				return new Api_Error_Response('Table '.$set->get_table().' isn\'t available through the API', 400);
			}


			switch($method) {
				case 'count':
					$result = array( 'count' => count($set) );
					break;
				case 'query':
					$class = get_class($pool);
					$all_columns = $class::get_all_columns_list();

					$extra_cols = array_diff($this->columns, $all_columns);

					if (count($extra_cols) > 0) {
						return new Api_Error_Response("The following columns doesn't exist: " . implode(', ', $extra_cols), 400);
					}

					foreach($set->it(empty($this->columns) ? false : $this->columns, $this->sort, $this->limit, $this->offset) as $obj) {
						$result[] = $obj->export();
					}
					break;
			}
		} catch(LSFilterException $e) {
			$query = $e->get_query();
			$position = $e->get_position();
			$full_error = $e->getMessage() . ': ' . substr($query,0,$position) . ' >>syntax error here<< ' . substr($query,$position);
			return new Api_Error_Response('Syntax error in query', 400, $full_error);
		} catch(Api_Error_Response $e) {
			// ObjectPool_Model might throw something (I actually
			// have no idea, it's interface is not easy to find out
			// more about) but we must check the resulting set's
			// mayi resource, and *that* might throw this exception
			// so.. just return it and be done with it.
			return $e;
		} catch(Exception $e) {
			return new Api_Error_Response($e->getMessage());
		}

		return new Api_Response(
			array(
				'body' => $result,
				'menu' => $this->get_menu()
			)
		);
	}

	/**
	 * Set parameters necessary to continue
	 *
	 * @param $params array
	 * @return array of errors, so if(!set_parameters($r)) print 'Everything A-OK';, C style
	 */
	private function set_parameters(array $params) {
		$errors = array();

		// limit
		if(!isset($params['limit'])) {
			$this->limit = self::DEFAULT_LIMIT; // @TODO I really want to display link to resources
			// without a set 'limit'
		} else if(!is_numeric($params['limit']) || (int) $params['limit'] != $params['limit'] || $params['limit'] < 1) {
			$errors[] = "The parameter 'limit' must be a positive number";
		} else {
			$this->limit = $params['limit'];
		}
		unset($params['limit']);

		// query
		if(!isset($params['query'])) {
			$errors[] = "Missing the parameter 'query'";
		} else {
			$this->query = $params['query'];
		}
		unset($params['query']);

		// offset
		if(isset($params['offset'])) {
			if(!is_numeric($params['offset']) || (int) $params['offset'] != $params['offset'] || $params['offset'] < 1) {
				$errors[] = "The parameter 'offset' must be a positive number";
			} else {
				$this->offset = $params['offset'];
			}
		}
		unset($params['offset']);

		// columns
		if(isset($params['columns'])) {
			if(is_array($params['columns'])) {
				$errors[] = "The parameter 'columns' must be a (comma separated) string";
			} else {
				$this->columns = explode(',', $params['columns']);
			}
		}
		unset($params['columns']);

		// sort
		if(isset($params['sort'])) {
			if(is_array($params['sort'])) {
				$errors[] = "The parameter 'sort' must be a (comma separated) string";
			} else {
				$this->sort = explode(',', $params['sort']);
			}
		}
		unset($params['sort']);

		if($params) {
			$errors[] = "Unknown parameters provided: ".implode(', ', array_keys($params));
		}

		return $errors;
	}
}
