<?php
require_once('request.php');
require_once('response.php');
require_once('error_response.php');
require_once('mayi_http_api.php');

require_once('op5/mayi.php');
require_once('op5/auth/Auth.php');

$nacoma_path = '/opt/monitor/op5/nacoma/';
define('NINJA_PATH', '/opt/monitor/op5/ninja/');
$request = Api_Request::factory();

// never send 200 OK on fatal errors..
// see https://bugs.php.net/bug.php?id=50921
register_shutdown_function(function() {
	if(PHP_SAPI == 'cli') {
		return;
	}
	$error = error_get_last();
	if ($error['type'] == E_ERROR) {
		header('HTTP/1.1 500 Internal Server Error');
	}
});

assert_options(ASSERT_CALLBACK, function($file, $line, $code, $desc = null) {
	throw new Api_Error_Response('Internal Server Error', 500, 'Please check ssl_error_log for more details');
});

/**
 * Setup kohana ('s autoload, really) so that we're not completely
 * without a DAL and helper functions.
 *
 * It also runs all hooks ("events") that Kohana has setup, most
 * notably the mayi ones.
 */
define('SKIP_KOHANA', true);
require_once rtrim(NINJA_PATH, '/').'/index.php';

/**
 * @return boolean
 */
function is_mobile() {
	return isset($_SERVER, $_SERVER['HTTP_USER_AGENT']) && preg_match('~op5-Monitor-Mobile-App/?.*~', $_SERVER['HTTP_USER_AGENT']);
}

/**
 * This is as good as it gets (which means: not very). A localhost request can
 * be spoofed (not sure how, but with modified "sent from" data), and the
 * request is also not considered to be a localhost_request if it's made to a
 * different peer, in which case the IPs will not match.
 *
 * @return boolean
 */
function is_local() {
	if(PHP_SAPI != 'cli') {
		$headers = getallheaders();
		// Allow testing to fake that we're not actually from localhost.
		// This should not have an impact on mayi since localhost should
		// *always* have more permissions than non-localhost originating
		// requests
		if(isset($headers['x-skip-localhost'])) {
			return false;
		}
	}
	return isset($_SERVER, $_SERVER['REMOTE_ADDR'], $_SERVER['SERVER_ADDR']) && $_SERVER['REMOTE_ADDR'] && $_SERVER['REMOTE_ADDR'] == $_SERVER['SERVER_ADDR'];
}

/**
 * @param $request Api_Request
 * @throws Api_Error_Response
 * @return Op5Auth
 */
function auth(Api_Request $request) {
	$auth_method = $request->get_auth_method();
	$conf = Op5Config::instance()->getConfig('http_api');
	if(!isset($conf['auth_method'][$auth_method])) {
		throw new Api_Error_Response('Invalid request', 400, "Bad value for 'auth_method' parameter (". htmlspecialchars($auth_method, ENT_QUOTES, 'utf-8')."), ".var_export($conf, true));
	}
	if(!$conf['auth_method'][$auth_method]) {
		// safety check: don't spawn more basic auth windows if it's impossible
		// to enter a combination to actually pass the auth; ergo, tell the user
		// that the API is offline
		foreach($conf['auth_method'] as $enabled) {
			if($enabled) {
				throw new Api_Error_Response('Disabled auth_method used', 412, "Given 'auth_method' (". htmlspecialchars($auth_method, ENT_QUOTES, 'utf-8').") is disabled through http_api.yml, enable it or use another 'auth_method'");
			}
		}
		throw new Api_Error_Response('Forbidden', 403, "The API is locked down, all 'auth_method's are disabled through http_api.yml");
	}
	switch($auth_method) {
	case 'ninja':
		Session::instance();
		$auth = Op5Auth::instance();
		if($auth->get_user() instanceof User_NoAuth_Model) {
			throw new Api_Error_Response('Unauthorized', 401, "You need to login to access this page. You tried to reuse the Monitor session but you seem to be logged out of Monitor");
		}
		break;
	case 'basic_auth':
		/* Fetch auth object without session support */
		$auth = Op5Auth::instance(array('session_key' => false));

		/* Check for an apache user */
		if ($auth->get_user() instanceof User_NoAuth_Model) {
			/* Login every time - not to keep session */

			if (!isset($_SERVER['PHP_AUTH_USER']) ||
				!isset($_SERVER['PHP_AUTH_PW']) ||
				!$auth->login($_SERVER['PHP_AUTH_USER'], $_SERVER['PHP_AUTH_PW']))
			{
				throw new Api_Error_Response('Unauthorized', 401, 'You need to login to access this page');
			}
		}
		break;
	default:
		// only happens if there's a method specified in http_api.yml but there's no logic
		// above for that case. add that logic
		throw new Api_Error_Response('Invalid request', 400, "Bad value for 'auth_method' parameter");
	}
	return $auth;
}

try {
	if($request->is_help() && $request->get_method() != 'GET') {
		throw new Api_Error_Response('Method not allowed', 405, 'You should use GET to request help pages');
	}
	$auth = auth($request);
	$mayi = op5mayi::instance();
	$mayi->act_upon(new mayi_http_api_constraints());
	$request->set_auth($auth);

	// Let's also disallow accessing beta features without the beta in
	// /api/*beta*/section, and do this independently so that the feature
	// itself does not have to bother checking for it
	if(!$request->is_beta() && array_key_exists($request->get_project_name(), op5config::instance()->getConfig('http_api.beta_features'))) {
		throw new Api_Error_Response('Not Found', 404, "You are trying to access a beta feature through the wrong URL. Beta features are available at /api/beta");
	}

	// GET-only endpoints are mayi-checked here, the other endpoints
	// are responsible for calling mayi themselves.
	switch($request->get_project_name()) {
	case 'command':
		require_once(dirname(__FILE__).'/../command.php');
		$response = generate_response($request);
		break;
	case 'config':
		require_once($nacoma_path.'api/http.php');
		$response = generate_response($request);
		break;
	case 'filter':
		require_once(dirname(__FILE__).'/../filter.php');
		$response = generate_response($request, $auth);
		break;
	case 'status':
		require_once(dirname(__FILE__).'/../livestatus.php');
		$response = generate_response($request, $auth);
		break;
	case 'report':
		mayi_run_api('monitor.monitoring.hosts:read.%s.report');
		require_once(dirname(__FILE__).'/../report.php');
		$response = generate_response($request, $auth);
		break;
	case 'perfdata':
		mayi_run_api('monitor.monitoring.hosts.perfdata:read.%s.status');
		require_once(dirname(__FILE__).'/../perfdata.php');
		$response = generate_response($request, $auth);
		break;
	case 'license':
		require_once(dirname(__FILE__).'/../license.php');
		$response = generate_license_response($request);
		break;
	case 'nachos':
		require_once(dirname(__FILE__).'/../nachos.php');
		$response = generate_nachos_response($request);
		break;
	case 'magellan':
		require_once(dirname(__FILE__).'/../magellan.php');
		$response = generate_magellan_response($request);
		break;
	case 'help':
	case '':
		$response = new Api_Response('<h1>ITRS OP5 Monitor API</h1><p>This is the ITRS OP5 Monitor HTTP API. Each Nagios object can be queried by using the following documentation.</p>', 200);
		break;
	default:
		$response = Api_Error_Response::with_usage('Invalid request', 400, $request);
	}
} catch (Api_Response $response) {
	if($response->getCode() == 401 && $request->get_auth_method() == 'basic_auth') {
		header('WWW-Authenticate: Basic realm="op5 Monitor API Access"');
	}
}
echo $response->send($request);
