<?php

require_once __DIR__.'/error_response.php';
require_once __DIR__.'/response.php';
require_once __DIR__.'/request.php';

/**
 * Base class for API endpoints that forward requests to other services.
 */
class ApiProxy {
	/**
	 * Forward request to another service.
	 *
	 * @param Api_Request $request The HTTP request we're responding to
	 * @param string $endpoint The base URL of the service we're forwarding to
	 * @throws Api_Error_Response
	 * @returns Api_Response
	 */
	public function forward_request(Api_Request $request, $endpoint) {
		$endpoint .= implode('/', $request->get_path());
		$params = $request->get_params();
		if (sizeof($params) > 0) {
		    $endpoint .= '?' . http_build_query($params);
		}
		$response = $this->curl_request(
		    $request,
		    $endpoint,
		    file_get_contents('php://input')
		);

		if ($response['http_code'] == 0) {
			throw new Api_Error_Response(
				'Bad Gateway',
				502,
				"Received no response from $endpoint. Is the service running?"
			);
		}

		return new ApiProxyResponse(
			$response['response'],
			$response['http_code'],
			$response['content_type']
		);
	}

	public function curl_request(Api_Request $request, $endpoint, $data) {
		$options = array(
			CURLOPT_CUSTOMREQUEST => $request->get_method(),
			CURLOPT_RETURNTRANSFER => true,
			CURLOPT_HTTPHEADER => $this->get_curl_headers(),
			CURLOPT_SSL_VERIFYPEER => false
		);
		$handle = $this->get_curl_handle($endpoint, $options);

		if (!empty($data)) {
			curl_setopt($handle, CURLOPT_POSTFIELDS, $data);
		}

		op5log::instance('http_api')->log(
			'debug',
			"Forwarding request to: $endpoint"
		);

		$response = curl_exec($handle);
		$http_code = curl_getinfo($handle, CURLINFO_HTTP_CODE);
		$content_type = curl_getinfo($handle, CURLINFO_CONTENT_TYPE);

		curl_close($handle);

		op5log::instance('http_api')->log(
			'debug',
			"Received response from $endpoint with status code $http_code"
		);

		return array(
			'response' => $response,
			'http_code' => $http_code,
			'content_type' => $content_type
		);
	}

	private function get_curl_handle($endpoint, $options) {
		$handle = curl_init($endpoint);

		foreach ($options as $k => $v) {
			curl_setopt($handle, $k, $v);
		}

		return $handle;
	}

	private function get_curl_headers() {
		$headers = array();

		if (function_exists('getallheaders')) {
			foreach (getallheaders() as $k => $v) {
				$headers[] = "$k: $v";
			}
		}
		else {
			$headers[] = "Accept: application/json";
		}

		return $headers;
	}
}

/**
 * Response class for proxied requests.
 */
class ApiProxyResponse extends Api_Response {
	/**
	 * Constructor
	 * @param $body Response body
	 * @param $code HTTP response code
	 * @param $content_type Content-Type header
	 */
	function __construct($body, $code = 200, $content_type = 'application/json')
	{
		$this->content_type = $content_type;
		parent::__construct($body, $code);
	}

	/**
	 * Send a response
	 *
	 * This sends a simplified response from a proxied request, where the
	 * body content is unchanged and the Content-Type header forwarded.
	 *
	 * @param $request Api_Request = null
	 * @return string
	 */
	public function send(Api_Request $request=null)
	{
		header("HTTP/1.0 ".$this->code." ".$this->status_codes[$this->code]);
		header("Content-Type: ".$this->content_type);
		return $this->body;
	}
}
