<?php

/**
 * A helper for dealing with common request conventions
 *
 * You should not have to deal with HTTP or lower-level PHP functions
 * directly.
 */
class Api_Request {
	private $beta = false;
	private $help = false;
	private $path = array();
	protected $params = array();
	private $full_path = '';
	private $project = false;
	private $output_format = 'json';
	private $auth = false;
	private $auth_method = 'basic_auth';

	public function __construct($request_uri)
	{
		/**
		 * Since PHP 5.4.7 leading double-slash "//api/config" results in the
		 * slashes being interpreted as the protocol separator in parse_url.
		 * Hence the first part of the path is assigned to the host property of
		 * the returned object.
		 */
		$request_uri = "/" . ltrim($request_uri, '/');
		$uri_components = parse_url($request_uri);
		if(isset($uri_components['query'])) {
			parse_str($uri_components['query'], $this->params);
			// rip out options that are used by the API framework
			// but can't be passed to the backend
			// (for example, report options chokes on invalid options,
			// and all of the other should as well, but don't atm)
			unset($this->params['format']);
			if(isset($this->params['auth_method'])) {
				$this->auth_method = $this->params['auth_method'];
				unset($this->params['auth_method']);
			}
		}
		$this->full_path = $uri_components['path'];
		$path = trim($uri_components['path'], '/');
		$path = explode('/', $path);
		$path = array_map('urldecode', array_map('urldecode', $path));
		array_shift($path); // /api
		if(reset($path) == 'help') {
			$this->help = true;
			array_shift($path); // /help requested
		}
		if(reset($path) == 'beta') {
			$this->beta = true;
			array_shift($path); // /beta requested
		}
		$this->output_format = $this->_detect_format();
		$this->project = array_shift($path); // /<project>
		$this->path = $path;
	}

	private function _detect_format() {
		if($this->is_help()) {
			// we should default to html since the help function is
			// accessed through a browser or perhaps fetched via wget,
			// either way it should be viewed through a browser
			return 'html';
		}
		if(isset($_GET['format'])) {
			switch($_GET['format']) {
				case 'json':
				case 'html':
				case 'xml':
					return $_GET['format'];
			}
		}
		if(isset($_SERVER['HTTP_ACCEPT'])) {
			foreach(explode(',', $_SERVER['HTTP_ACCEPT']) as $accepted) {
				if(in_array($accepted, array(
				'application/json',
				'application/xml',
				'text/html'
				))) {
					return preg_replace('~(application|text)/~', null, $accepted);
				}
			}
		}
		if(isset($_SERVER['CONTENT_TYPE'])) {
			switch($_SERVER['CONTENT_TYPE']) {
				case 'application/json':
				case 'application/xml':
				case 'text/html':
					return preg_replace('~(application|text)/~', null, $_SERVER['CONTENT_TYPE']);
			}
		}
		return 'json';
	}

	static function factory(array $options = array()) {
		$props_needed = array(
			'request_uri' => isset($_SERVER['REQUEST_URI']) ? $_SERVER['REQUEST_URI'] : null
		);
		foreach($props_needed as $key => $default) {
			if(isset($options[$key])) {
				$props_needed[$key] = $options[$key];
			}
		}
		return new self($props_needed['request_uri']);
	}

	/**
	 * @param $auth op5auth
	 */
	public function set_auth(op5auth $auth) {
		$this->auth = $auth;
	}

	/**
	 * @return string
	 */
	public function get_auth_method() {
		return $this->auth_method;
	}

	/**
	 * @return string
	 */
	public function get_project_name()
	{
		return $this->project;
	}

	/**
	 * @return string GET, POST, DELETE, PATCH, PUT
	 */
	public function get_method() {
		return $_SERVER['REQUEST_METHOD'];
	}

	/**
	 * @return string xml|json|html
	 */
	public function get_output_format()
	{
		return $this->output_format;
	}

	/**
	 * Get the request path, relative to the project's 'mount point'
	 *
	 * @return array, e.g. [ 'host', 'monitor' ]; higher index = more granularity
	 */
	public function get_path()
	{
		return $this->path;
	}

	/**
	 * Get a map of request GET parameters
	 *
	 * @return array
	 */
	public function get_params()
	{
		return $this->params;
	}

	/**
	 * Get a single GET parameter, or the default value if the parameter
	 * was left out
	 *
	 * @param $name string
	 * @param $default string = null
	 * @return string
	 */
	public function get_param($name, $default = null) {
		if(isset($this->params[$name])) {
			return $this->params[$name];
		}
		return $default;
	}

	/**
	 * Get a deserialized version of the request body (POST data)
	 *
	 * @return string
	 */
	public function get_body()
	{
		$body = file_get_contents("php://input");
		if (empty($body)) {
			return null;
		}

		switch($this->get_content_type()) {
			case 'application/xml':
				$result = $this->xml_decode($body);
				if(null === $result) {
					throw new Exception("Invalid xml provided");
				}
				return $result;
			case 'application/json':
				$result = json_decode($body, true);
				if(null === $result) {
					throw new Exception("Invalid json provided");
				}
				return $result;
			case 'application/octet-stream':
				return $body;
			default:
				// This is a weird implementation in our API,
				// not changing it now because of possible
				// backwards compatibility issues.
				//
				// > If a Content-Type header field is not
				// > present, the recipient MAY either assume a
				// > media type of "application/octet-stream"
				// > ([RFC2046], Section 4.5.1) or examine the
				// > data to determine its type.
				//
				// https://tools.ietf.org/html/rfc7231#section-3.1.1.5
				parse_str($body, $body);
		}
		return $body;
	}

	/**
	 * Which content type did the request request? :P
	 *
	 * @return string|NULL
	 */
	public function get_content_type() {
		if (isset($_SERVER['CONTENT_TYPE'])) {
			$content_type = explode(";", $_SERVER['CONTENT_TYPE']);
			return $content_type[0];
		}
		return NULL;
	}

	/**
	 * The user wants to access something that's categorized as a beta
	 * feature and thus opt-in.
	 *
	 * @return boolean
	 */
	function is_beta() {
		return $this->beta;
	}

	/**
	 * Check if this is a help document or not
	 * @return True if yet, false if no
	 */
	public function is_help()
	{
		return $this->help;
	}

	/**
	 * Get the current user's username
	 */
	public function get_username()
	{
		if ($this->auth)
			return $this->auth->get_user()->get_username();
		return "Unknown user";
	}

	/**
	 * Use this and safely ignore whether to link to https or not,
	 * ip or hostname, etc.
	 *
	 * @return string
	 */
	public function get_base_url()
	{
		static $base_url;
		if(!$base_url) {
			$base_url = 'http'.(!empty($_SERVER['HTTPS'])?'s':'').'://'.(isset($_SERVER['HTTP_HOST'])?$_SERVER['HTTP_HOST']:$_SERVER['SERVER_NAME']);
		}
		return $base_url;
	}

	/**
	 * Used in conjunction with @see get_base_url() but more complete
	 *
	 * @param $link_to_format string = null
	 * @return string
	 */
	public function get_full_url($link_to_format = null)
	{
		$full_url = $this->get_base_url().$this->full_path;
		if(!$link_to_format) {
			if($this->params) {
				$full_url .= '?'.http_build_query($this->params);
			}
			return $full_url;
		}

		if(!in_array($link_to_format, array('json', 'html', 'xml'))) {
			throw new Exception("Can't format link for the format '$link_to_format'");
		}
		return $full_url.'?'.http_build_query(array('format' => $link_to_format) + $this->params);
	}

	private function xml_decode_worker($parent)
	{
		$val = false;
		if ($parent->nodeType === XML_TEXT_NODE)
			return $parent->nodeValue;

		if ($parent->nodeType !== XML_ELEMENT_NODE)
			return;

		foreach ($parent->childNodes as $node) {
			if ($node->nodeType === XML_ELEMENT_NODE) {
				if ($node->hasAttribute('name'))
					$val[$node->getAttribute('name')] = $this->xml_decode_worker($node);
				else
					$val[] = $this->xml_decode_worker($node);
			}
			if ($node->nodeType === XML_TEXT_NODE && !$node->isWhitespaceInElementContent())
			{
				$val = $node->nodeValue;
			}
		}
		return $val;
	}

	private function xml_decode($body)
	{
		$out = array();
		$doc = new DOMDocument();
		@$doc->loadXML($body);
		if ($doc === false)
			return null;
		$root = $doc->documentElement;
		if ($root == null)
			return null;
		foreach ($root->childNodes as $node) {
			if ($node->nodeType !== XML_ELEMENT_NODE)
				continue;
			$out[$node->getAttribute('name')] = $this->xml_decode_worker($node);
		}
		return $out;
	}
}
