<?php

/**
 * API response class
 *
 * This class is responsible for formatting and sending a
 * response back to the client.
 *
 * Extends Exception, to allow us to take a short-cut out of deep call
 * hierarchies. This is really for Api_Error_Response's benefit, since
 * PHP neither supports multiple inheritance, or has an exception interface.
 */
class Api_Response extends Exception
{
	const DEFAULT_STATUS_CODE = 500;

	protected $type = false; /**< The object type we refer to */
	protected $title = false; /**< Something pretty and describing, usually object name or type */
	protected $code; /**< HTTP status code */
	protected $status_codes = array(
		'200' => "OK",
		'201' => "Created",
		'202' => "Accepted",
		'400' => "Bad request",
		'401' => "Unauthorized",
		'403' => "Forbidden",
		'404' => "Not Found",
		'405' => "Method Not Allowed",
		'406' => "Not Acceptable",
		'409' => "Conflict",
		'412' => "Precondition Failed",
		'415' => "Unsupported Media Type",
		'500' => "Internal Server Error",
		'501' => "Not Implemented",
		'502' => "Bad Gateway",
		'503' => "Service Unavailable"
	);

	/** array [id] => [title => string, content => string] */
	protected $troubleshooting = array(
		'ssl' => array(
			'title' => "SSL certificate is invalid",
			'content' => "<p>By default, op5 Monitor ships with an unsigned SSL certificate. If you do not replace this with a signed certificate, the most HTTP access methods (including the ones given in the examples) will deny your access. This is solved the following way:</p>
				<p>cURL: Add a <code>--insecure</code> flag:</p>
				<div class='curl'>
					<pre><code>curl &lt;API endpoint here&gt; <strong>--insecure</strong> -u '&lt;your username&gt;:&lt;your password&gt;'</code></pre>
				</div>

				<p>Python: Adding a <code>verify=False</code> flag:</p>
				<div class='python'>
					<pre><code>requests.get('&lt;API endpoint here&gt;', auth=('&lt;your username&gt;', '&lt;your password&gt;'), <strong>verify=False</strong>)</code></pre>
				</div>

				<p>PHP with cURL: Adding a <code>CURLOPT_SSL_VERIFYPEER</code> flag:</p>
				<div class='php'>
					<pre><code>curl_setopt(\$a_handle, <strong>CURLOPT_SSL_VERIFYPEER</strong>, false);</code></pre>
				</div>"
		),
		'connect' => array(
			'title' => "Can I connect to the API in another way?",
			'content' => "Yes, all you need is a way to perform a HTTP method in your language. Study the minimal examples and take note of which HTTP method is used, which URI that's used and what data is sent to the URI."
		),
		'method' => array(
			'title' => "What's with PUT, DELETE, GET..?",
			'content' => "These are HTTP methods. There are examples of each method on the object's page."
		),
		'status_code' => array(
			'title' => "HTTP status codes used by Monitor",
			'content' => "<p>The HTTP response status codes are quite general and Monitor tries to follow the formal descriptions in every possible situation. Therefore, Wikipedia's article <a href='http://en.wikipedia.org/wiki/HTTP_status_code'>List of HTTP status codes</a> should help you get back on track.</p><p>These are the status codes that Monitor uses:</p>%s<p>Note that the text might change on a case-by-case basis, and that the numerical status code takes precedence.</p>"
		),
		'cant_find' => array(
			'title' => "Why can't the API find my object?",
			'content' => "First, login to the Monitor GUI and see if you can access it there. If you can, make sure you ask for it correctly by urlencoding the object's name twice. For example, if your host is called <em>my/host</em>, a call for <var>/api/status/my/host/a_service</var> need to be this instead: <var>/api/status/my%252Fhost/a_service</var>"
		),
		'login_custom_auth' => array(
			'title' => "How do I choose how to login with LDAP or locally?",
			'content' => '<p>When you have more than one auth module, for example "Default" and "LDAP", you need to specify which to authenticate against. This is done with the dollar character (\'$\').</p>

			<p>Thus, this regular call:<br />
			<var>curl -u user:password https://monitor/api/status/host</var></p>

			<p>
				becomes<br/>
				<var>curl -u \'user$LDAP:password\' https://monitor/api/status/host</var><br/>
				or<br/>
				<var>curl -u \'user$Default:password\' https://monitor/api/status/host</var>
			</p>

			<p>Notice how the dollar sign (\'$\') needs quoting depending on the environment (in bash, it will always need to be quoted).</p>

			<p>The first way of calling the API:<br />
			<var>curl -u user:password https://monitor/api/status/host</var><br />
			will still work, provided that you want to authenticate against the default driver. The default driver can always be specified from within the GUI: Configure -> Auth Modules.</p>'
		),
		'restricting_access' => array(
			'title' => 'Restricting access to the API',
			'content' => '<p>By editing /etc/op5/http_api.yml, you can decide wether to accept or reject connections to the API based on connection method. For now, there are two categories: <em>basic_auth</em> and <em>ninja</em>.</p>
				<ul>
					<li>If you disallow <em>basic_auth</em>, you can no longer access the API through scripts as described in all examples in this documentation.</li>
					<li>If you disallow <em>ninja</em>, you can no longer access the API through AJAX requests carried out by code in Monitor, such as the <em>Host Wizard</em>.</li>
				</ul>'
		),
		'custom_variables' => array(
			'title' => "How do I handle custom variables?",
			'content' => "Custom variables are just key-values stored with each Nagios object, as such, the API treats them the same way. Just make sure that each key is prepended with an underscore (_), otherwise the request will be returned with an erroneous status code."
		),
		'templates' => array(
			'title' => "Why can I create objects without all required properties set? Or, How do templates work?",
			'content' => "<p>Most objects has templates, i.e. hosts have their <var>template</var> property. There is usually a default template which will be used. If that template has a set <var>file_id</var> for example, you are not required to include a <var>file_id</var> in the definition of your host, since it will be inherited from the template. You can either rely on the default template, or provide another value of <var>template</var>.</p><p>For further configuration on the matter, please see the configuration tool inside the Monitor GUI, and read about <strong>Templates</strong> in the op5 Monitor Administrator Manual.</p>"
		),
		'resource' => array(
			'title' => "What is a 'resource'?",
			'content' => "<p>When listing objects, you are given an additional property called <em>resource</em>. This let's you know the resource's location to which you might send a POST or PUT request. Do note that the resource is not permanent, since a removed/renamed object will make the URI cease to exist.</p>
				<p>The resource is urlencoded so that it is always usable in its current form.</p>"
		),
		'invalid_request' => array(
			'title' => "Why do I get the \"Couldn't save object\" error response?",
			'content' => "Make sure that you
			<ul>
				<li>Include the correct content header (<var>content-type: application/json</var> for example)</li>
				<li>Check that your relations (the template you're using, the host you're placing the service on, the group you're putting your host in, etc) points at real target objects</li>
				<li>Submit <a href='http://jsonlint.com'>valid JSON</a> or XML</li>
			</ul>"
		),
		'handling_errors' => array(
			'title' => 'How are errors reported?',
			'content' => "<p>You should always check the return code of the response. If it is 400 or larger, an error occured. <em>Which error?</em> you might ask. To find out what went wrong, you need to check the response's body. It always consists of a datastructure with the keys <var>error</var> and <var>full_error</var>. Here's an example:</p>

<dl>
	<dt>Request</dt>
	<dd><var>curl -u user:password -d '&lt;root&gt;&lt;param name=\"host_name\"&gt;hello&lt;/host_name&gt;&lt;address&gt;good_bye&lt;/address&gt;&lt;/root&gt;' -H 'content-type: application/xml' https://monitor/api/config/host -i</var></dd>

	<dt>Response</dt>
	<dd><pre><var>HTTP/1.0 <strong>400</strong> Bad request
Date: Thu, 11 Jul 2013 13:42:17 GMT
Server: Apache/2.2.15
X-Powered-By: PHP/5.3.3
Set-Cookie: ninjasession=56fl5qktd8bsb9pl0r7uku7lp3; path=/
Expires: Thu, 19 Nov 1981 08:52:00 GMT
Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0
Pragma: no-cache
Content-Length: 135
Connection: close
Content-Type: application/xml; charset=utf-8

&lt;?xml version=\"1.0\"?&gt;
&lt;root&gt;
  &lt;param name=\"<strong>error</strong>\"&gt;Bad Request&lt;/param&gt;
  &lt;param name=\"<strong>full_error</strong>\"&gt;Invalid xml provided&lt;/param&gt;
&lt;/root&gt;</var></pre></dd></dl>
			  <p>Note that the <var>full_error</var> key does not have a fixed type, it can be a string, a list of strings, numbers, etc. Thus, it requires some sort of manipulation on your part.</p>"
		),
		'python_requests' => array(
			'title' => "Trying to use the python examples, getting 'ImportError: No module named requests'",
			'content' => "In the examples we use a python library called <strong><a href='http://docs.python-requests.org/' target='_blank'>requests</a></strong> which makes it easy to perform HTTP requests. <strong>requests</strong> is not installed together with python so you have to install it yourself. If you are using <a href='http://yum.baseurl.org/' title='A common package system' target='_blank'>yum</a>, <var>yum install python-requests</var> should be enough. Python savvy people might like to use <a href='http://www.pip-installer.org' target='_blank'>pip</a> instead: <var>pip install requests</var>, but the choice is up to you. We recommend you to read up on the subject if python is your preferred language."
		),
		'unknown' => array(
			'title' => "I am experiencing problems, where should I go next?",
			'content' => "You could contact your op5 Monitor technician or visit <a href='http://www.op5.com/support/'>http://www.op5.com/support/</a> to acquire more help."
		)
	);

	/**
	 * The content of these properties will be wrapped in <a href> in the HTML output
	 */
	public $properties_to_linkup = array(
		'resource',
		'next_resource'
	);
	/**
	 * Constructor
	 * @param $body Response body
	 * @param $code HTTP response code
	 */
	function __construct($body, $code = 200)
	{
		$this->body = $body;
		if(!array_key_exists($code, $this->status_codes)) {
			// Somebody provided a code that is not within the
			// scope of the HTTP API. That code needs to be added
			// to the Api_Response's knowledge, or replaced with a
			// code that the Api_Response know of. Until then, we
			// guess the code's intention and default to a "bad"
			// code because the common good codes a) are known to
			// Api_Response, and b) is the default argument to
			// this method, and that argument is obviously
			// overridden.
			if(is_numeric($code) && $code >= 200 && $code < 300) {
				$new_code = 200;
			} else {
				$new_code = self::DEFAULT_STATUS_CODE;
			}
			op5log::instance('http_api')->log('debug', "Tried to send '$code' as an unspecified HTTP API status code, overriding with '$new_code'");
			$code = $new_code;
			unset($new_code);
		}
		$this->code = $code;

		$codes = "<ul>\n";
		foreach($this->status_codes as $code => $description) {
			 $codes .= "<li>$code $description</li>\n";
		}
		$codes .= "</ul>";
		$this->troubleshooting['status_code']['content'] = sprintf($this->troubleshooting['status_code']['content'], $codes);
	}

	/**
	 * @param $properties array Including the keys: name (string), type (string),
	 * default (mixed), description (string), required (boolean)
	 * @param $title string = null
	 * @return string html
	 */
	protected function _format_properties($properties, $title = null)
	{
		$properties_as_html = '<table>
			<tr>
				<th>Parameter (<span class="required">required</span>)</th>
				<th>Type</th>
				<th>Default</th>
				<th>Description</th>
			</tr>';
		if($title) {
			$title = (string) $title;
			$properties_as_html = "<h2>Properties of <strong>$title</strong> resources</h2>".$properties_as_html;
		}
		foreach($properties as $property) {
			$required_class = 'class="required"';
			if(!$property['required']) {
				$required_class = '';
			}
			$properties_as_html .= "<tr $required_class>
				<td><var>".$property['name']."</var></td>
				<td>".$property['type']."</td>
				<td>".$property['default']."</td>
				<td>".$property['description']."</td>
			</tr>";
		}
		return $properties_as_html.'</table>';
	}

	/**
	 * @param $troubleshooting array
	 * @return array [menu => string, content => string]
	 */
	protected function _format_troubleshooting($troubleshooting)
	{
		$menu = null;
		$content = null;
		foreach($troubleshooting as $key => $item) {
			$menu .= "<li><a href='#trouble_$key'>{$item['title']}</a></li>\n";
			$content .= "<dt id='trouble_$key'>{$item['title']}</dt>
				<dd>{$item['content']}</dd>\n";
		}
		return array(
			'menu' => "<h3>Troubleshooting</h3>
				<ul>$menu</ul>",
			'content' => "<h3>Troubleshooting</h3>
				<dl>
					$content
				</dl>"
		);
	}

	/**
	 * Serialize output
	 *
	 * @param $output_format string
	 * @param $request Api_Request
	 * @return string
	 */
	protected function _get_adjusted_response($output_format, $request)
	{
		switch($output_format) {
			case 'xml':
				$content_type = 'application/xml';
				$body = $this->body;
				if(is_array($this->body) && isset($this->body['body'])) {
					$body = $this->body['body'];
				}
				$body = $this->xml_encode($body)->saveXML();
				break;
			case 'html':
				$content_type = 'text/html';
				$title = "ITRS OP5 Monitor API - $this->title";
				$menu_html = null;
				$description = null;
				$properties = null;
				$trouble_menu = null;
				$trouble_content = null;
				if(is_array($this->body) && isset($this->body['menu']) && isset($this->body['body'])) {
					$body = null;
					if($this->body['body']) {
						// allow an empty body
						$body = $this->html_encode($this->body['body']);
					}
					if(!is_array($this->body['menu']) || isset($this->body['menu']['title'])) {
						// 'title' should be set in every meny, but we want body['menu']
						// to be an array of menu-arrays. Crystal clear?
						$this->body['menu'] = array($this->body['menu']);
					}
					foreach($this->body['menu'] as $menu) {
						$menu_html .= "<h2>".$menu['title']."</h2><ul>";
						foreach($menu['items'] as $name => $menu_item) {
							if($name === 'index') {
								// sections called index are aliases to one-level-up
								continue;
							}
							$menu_html .= sprintf(
								"<li><a href='%s'>%s</a></li>",
								$menu_item['resource'],
								$name
							);
						}
						$menu_html .= "</ul>";
					}
					if(isset($this->body['description']) && $this->body['description']) {
						$description = $this->body['description'];
					}
					if(isset($this->body['properties']) && $this->body['properties']) {
						$properties = $this->_format_properties($this->body['properties'], $this->title);
					}
					$troubleshooting = $this->troubleshooting;
					if(isset($this->body['troubleshooting']) && $this->body['troubleshooting']) {
						$troubleshooting = array_merge($this->body['troubleshooting'], $troubleshooting);
					}
					$troubles = $this->_format_troubleshooting($troubleshooting);
					$trouble_menu = $troubles['menu'];
					$trouble_content = $troubles['content'];
				} else {
					$body = $this->html_encode($this->body);
				}
				if(0 !== strpos($body, '<!DOCTYPE')) {
					$links_to_other_formats = null;
					if(!$request->is_help() && $request->get_path()) {
						// "path" being set means that we got both a "project"
						// (status/config et al.) and a sub section such as "host",
						// since we don't want to show the "alternative xml view"
						// of the help's index page (at least not until we're doing
						// HATEOAS)
						$xml_link = $request->get_full_url('xml');
						$json_link = $request->get_full_url('json');

						$links_to_other_formats = "<div class='meta'>
							<p>You are viewing the HTML version of this document.</p>
							<p>If you want to make calls to the API you should use either the 
								<a href='" . htmlspecialchars($json_link, ENT_QUOTES, 'UTF-8') . "'>JSON</a> or 
								<a href='" . htmlspecialchars($xml_link, ENT_QUOTES, 'UTF-8') . "'>XML</a> versions of this page.
							</p>
						</div>";
					}
					$body = "<!DOCTYPE html>
					<html><head>
					<link type='text/css' rel='stylesheet' href='/api/css/reset.css' />
					<link type='text/css' rel='stylesheet' href='/api/css/monitor.css' />
					<link rel='index' href='/api' />
					<title>$title</title>
					</head><body>
					<div id='container'>
						<div id='sidebar'>
							<div id='monitor'>op5 Monitor</div>
							<p>HTTP API</p>
							<ul>
								<li><a href='/api/help/command'>Command</a></li>
								<li><a href='/api/help/config'>Config</a></li>
								<li><a href='/api/help/report'>Report</a></li>
								<li><a href='/api/help/status'>Status</a></li>
								<li><a href='/api/help/filter'>Filter</a></li>
								<li><a href='/api/help/magellan'>Magellan</a></li>
								<li><a href='/api/help/beta/perfdata'>Perfdata</a> (beta)</li>
								<li><a href='/api/help/beta/license'>License</a> (beta)</li>
							</ul>
							<p>This is the ITRS OP5 Monitor HTTP API. Each Nagios object can be queried by using the following documentation.</p>
							$menu_html
							$trouble_menu
						</div>
						<div id='main'>
							<h1>$this->title</h1>
							$description
							$properties
							$links_to_other_formats
							$body
							$trouble_content
							<h2>Getting help</h2>
							<p>Please visit <a href='/api/help/command'>/api/help/command</a>, <a href='/api/help/config'>/api/help/config</a>, <a href='/api/help/report'>/api/help/report</a>, <a href='/api/help/status'>/api/help/status</a> or <a href='/api/help/filter'>/api/help/filter</a>.</p>
						</div>
					</div>
					</body></html>";
				}
				break;
			case 'json':
			default:
				$body = $this->body;
				if(is_array($this->body) && isset($this->body['body'])) {
					$body = $this->body['body'];
				}
				$content_type = 'application/json';
				$body = json_encode($body);
				break;
		}
		return $body;
	}

	/**
	 * This is the contents of the response. It needs to be formatted in
	 * JSON or another output format, before returned to the user.
	 *
	 * @return array
	 */
	public function get_body() {
		return $this->body;
	}

	/**
	 * @return int
	 */
	public function get_code() {
		return $this->code;
	}

	/**
	 * Send a response
	 *
	 * This function takes a request object to find output format from, but
	 * in case of errors parsing the input, it needs to be optional.
	 *
	 * @param $request Api_Request = null
	 * @return string
	 */
	public function send(Api_Request $request=null)
	{
		$output_format = false;
		if ($request) {
			$output_format = $request->get_output_format();
			$path = $request->get_path();
			$this->type = reset($path);
			$this->title = end($path);
		}

		header("HTTP/1.0 ".$this->code." ".$this->status_codes[$this->code]);
		switch($output_format) {
			case 'xml':
				$content_type = 'application/xml';
				break;
			case 'html':
				$content_type = 'text/html';
				break;
			case 'json':
			default:
				$content_type = 'application/json';
				break;
		}
		header("Content-Type: $content_type; charset=utf-8");
		return $this->_get_adjusted_response($output_format, $request);
	}

	private function xml_encode_worker($k, $v, $parent, $doc, $elem_name="value") {
		if (!is_numeric($k) || preg_match('/\d{10}/', $k))
			$parent->setAttribute('name', $k);
		if (is_object($v)) {
			$v = get_object_vars($v);
		}
		if (is_array($v)) {
			foreach ($v as $vk => $vv) {
				$value = $doc->createElement($elem_name);
				$parent->appendChild($value);
				$this->xml_encode_worker($vk, $vv, $value, $doc, "value");
			}
		}
		else {
			// regex found at http://stackoverflow.com/a/3466609/49879
			$v = preg_replace('/[^\x{0009}\x{000a}\x{000d}\x{0020}-\x{D7FF}\x{E000}-\x{FFFD}]+/u', '', $v);
			$parent->appendChild($doc->createTextNode($v));
		}
	}

	/**
	 * We need to do the seemingly annoying<key>$key</key><value>$value</value>
	 * encoding, because timeperiod exceptions and variables such as 2dcoords
	 * are not valid as XML keys
	 */
	private function xml_encode($data, $root_name='root') {
		$doc = new DOMDocument();
		$doc->formatOutput = true;
		$root = $doc->createElement($root_name);
		$doc->appendChild($root);
		$this->xml_encode_worker(0, $data, $root, $doc, "param");
		return $doc;
	}

	/**
	 * @param $data mixed
	 * @param $escape boolean = false, not used for the whole body but used when listing objects' properties
	 */
	private function html_encode($data, $escape = false) {
		if(is_bool($data)) {
			return (int) $data;
		}
		if(is_string($data)) {
			if (empty($data) && (!is_string($data) || strval($data) == ""))
				return '[empty]';
			if($escape) {
				return htmlspecialchars($data);
			}
			return $data;
		}
		if(is_numeric($data)) {
			return (string) $data;
		}
		$string = null;
		if(is_array($data)) {
			if(empty($data)) {
				return '[empty]';
			}
			if (array_keys($data) !== range(0, count($data) - 1)) {
				// Associative map/hashmap/dictionary (copied from Zend's json encoding)
				$string .= '<dl>';
				foreach($data as $key => $value) {
					if(in_array($key, $this->properties_to_linkup) && is_string($value)) {
						$string .= "<dt>$key</dt><dd><a href='$value'>".htmlspecialchars($value)."</a></dd>";
					} else {
						$string .= "<dt>".htmlspecialchars($key)."</dt><dd>".$this->html_encode($value, true)."</dd>";
					}
				}
				$string .= '</dl>';
			} else {
				// Array/list
				$string .= '<ul>';
				foreach($data as $value) {
					$string .= "<li>".$this->html_encode($value, true)."</li>";
				}
				$string .= '</ul>';
			}
		}
		if($data instanceof stdClass) {
			return $this->html_encode((array) $data);
		}
		return $string;
	}
}
