<?php

require_once __DIR__.'/lib/error_response.php';
require_once __DIR__.'/lib/op5mayi_wrapper.php';
require_once __DIR__.'/lib/op5license_wrapper.php';
require_once __DIR__.'/lib/op5log_wrapper.php';
require_once __DIR__.'/interface/logger_interface.php';

/**
 * @param $request HTTP request
 * @return Api_Response
 */
function generate_license_response($request) {
	$mayi = new Op5MayIWrapper();
	$path = $request->get_path();
	$lic_path = '/etc/op5license/op5license.lic';
	$logger = new Op5LogWrapper(op5log::instance('http_api'));

	switch ($request->get_method()) {

		case 'PUT':
			if ($mayi->run("monitor.license:create", array(), $messages) == false) {
				return new Api_Response('Forbidden', 403);
			}
			if(!$request->is_help() && !empty($path) && $path[0] == "data") {
				if($request->get_content_type() != 'application/octet-stream') {
					return new Api_Error_Response(
						"Unsupported Media Type", 415,
						"The request must specify its ".
						"content-type as application/octet-stream"
					);
				}

				$content = $request->get_body();
				if(!$content) {
					return new Api_Error_Response(
						"Bad Request",
						400,
						"Missing file"
					);
				}

				/*
				 * Verify the license in a temporary file before installing
				 * it to the correct place
				 */
				$temp_file = tempnam(__DIR__, "temp_lic");
				if($temp_file === false || file_put_contents($temp_file, $content) === false) {
					@unlink($temp_file);
					$logger->log("error", "could not create file on disk: $temp_file");
					return new Api_Error_Response(
						"Internal Server Error",
						500,
						"Check /var/log/op5/http_api.log for more details"
					);
				}
				$temp_lic = new Op5LicenseWrapper(new op5license($temp_file));
				$valid = $temp_lic->is_valid();
				@unlink($temp_file);
				if ($valid) {
					try {
						$lic_file = new SplFileObject($lic_path, 'w');
						return License::install($content, $lic_file, $logger);
					} catch (Exception $e) {
						$logger->log("error", "could not open $lic_path for writing. Please make sure that it's owned by apache:apache");
						return new Api_Error_Response(
							"Internal Server Error",
							500,
							"Check /var/log/op5/http_api.log for more details"
						);
					}
				} else {
					return new Api_Error_Response(
						"Bad Request", 400,
						"The uploaded file does not seem to be a valid ".
						"license file. Please contact OP5 Support if ".
						"you believe this is an error"
					);
				}
			} else if(empty($path)) {
				return new Api_Error_Response("Method Not Allowed", 405);
			} else {
				return new Api_Error_Response("Not Found", 404);
			}
			break;

		case 'GET':
			if ($mayi->run("monitor.license:read", array(), $messages) == false) {
			    return new Api_Response('Forbidden', 403);
			}
			if($request->is_help() && !empty($path) && $path[0] == "data") {
				return License::data_resource_help($request->get_base_url());
			} else if(empty($path)) {
				return License::resource_help($request->get_base_url());
			} else if($request->is_help()) {
				return new Api_Error_Response("Not Found", 404);
			} else if(!empty($path) && $path[0] == "data") {
				return License::view(
					$mayi, 
					new Op5LicenseWrapper(new op5license('/etc/op5license/op5license.lic'))
				);
			} else {
				return new Api_Error_Response("Not Found", 404);
			}
			break;

		default:
			return new Api_Error_Response("Method Not Allowed", 405);
		}

}

class License {

	/**
	 * Will return the metadata from the license provided in $license as an API
	 * response
	 *
	 * @param PermissionsInterface $mayi A mayi permission object
	 * @param LicenseInterface $license An op5license object (thus, a parsed license)
	 * @return Api_Response A JSON data structure with with metadata of the license,
	 *	the license expiry date and a 200 OK response code
	 * @throws Api_Error_Response
	 */
	public static function view(PermissionsInterface $mayi, LicenseInterface $license) {
		$metadata = $license->get_metadata();
		$mayi->run('monitor.monitoring.hosts:read', array(), $messages, $metrics);
		$metric_time = $metrics['license_time'];

		return new Api_Response(
			array('metadata' => $metadata, 'expires_at' => $metric_time->get_max()),
			200
		);
	}

	/**
	 * Write content to a file if you have permission to create a license
	 *
	 * @param string $content license content.
	 * @param SplFileObject $file the file to which we should install the license.
	 * @param LoggerInterface $logger logging backend
	 * @return Api_Response HTTP response.
	 */
	public static function install($content, $file, LoggerInterface $logger) {
		assert(gettype($content) == 'string');

		$written = $file->fwrite($content);
		if ($written == NULL || $written == 0) {
			$logger->log("error", "could not write to ".$file->getRealPath().". Please make sure that it's owned by apache:apache");
			return new Api_Error_Response(
				"Internal Server Error",
				500,
				"Check /var/log/op5/http_api.log for more details"
			);
		}

		$file->fflush();

		$content_len = strlen($content);
		assert($content_len == $written);

		return new Api_Response("License uploaded successfully", 200);
	}

	/**
	 * Return the license help page
	 *
	 * @param string $basurl The server base url e.g. 'localhost'
	 * @return Api_Response A HTTP response with help for the license endpoint
	 */
	public static function resource_help($baseurl) {
		assert(gettype($baseurl) == 'string');

		$file_url = $baseurl.'/api/help/beta/license/data';
		return new Api_Response(
			array(
				'menu' => License::get_menu(),
				'body' => '<h2>License</h2>'.
					"<p>Here you can <a href='$file_url'>manage ".
					'your license for OP5 Monitor</a>.'
			)
		);
		return new Api_Response($output);
	}

	/**
	 * Return the license/data help page
	 *
	 * @param string $basurl The server base url e.g. 'localhost'
	 * @return Api_Response A HTTP response with help for the license/data endpoint
	 */
	public static function data_resource_help($baseurl) {
		assert(gettype($baseurl) == 'string');

		ob_start();
		$license_endpoint_uri = $baseurl.'/api/beta/license/data';
		require_once(__DIR__.'/license/data.php');
		$body = ob_get_clean();
		return new Api_Response(
			array(
				'menu' => License::get_menu(),
				'body' => $body
			)
		);
	}

	private static function get_menu() {
		return array(
			array(
				'title' => 'License',
				'items' => array(
					'data' => array(
						'resource' => '/api/help/beta/license/data'
					),
				)
			)
		);
	}
}
