<?php

/**
 * @param $request Api_Request
 * @param $auth Op5Auth
 * @return Api_Response
 */
function generate_response(Api_Request $request, Op5Auth $auth) {
	$api = new Livestatus_Api($auth);
	return $api->handle_request($request);
}

/**
 * Livestatus communication handler, manages requests that explicitly queries
 * a certain object type: api/status/host
 */
class Livestatus_Api
{
	private $auth;
	protected $obj_types = array(
		'host' => array(
			'description' => 'A host definition is used to define a physical server, workstation, device, etc. that resides on your network.',
			'duplicate' => array(
				'obsess' => 'obsess_over_host'
			)
		),
		'service' => array(
			'description' => 'A service definition is used to identify a "service" that runs on a host. The term "service" is used very loosely. It can mean an actual service that runs on the host (POP, SMTP, HTTP, etc.) or some other type of metric associated with the host (response to a ping, number of logged in users, free disk space, etc.). The different arguments to a service definition are outlined below.',
			'duplicate' => array(
				'obsess' => 'obsess_over_service'
			)
		),
		'hostgroup' => array(
			'description' => 'A host group definition is used to group one or more hosts together for simplifying configuration with object tricks or display purposes in the Monitor web GUI.'
		),
		'servicegroup' => array(
			'description' => 'A service group definition is used to group one or more services together for simplifying configuration with object tricks or display purposes in the Monitor web GUI.'
		),
		'contact' => array(
			'description' => 'Your nagios contacts.'
		),
		'contactgroup' => array(
			'description' => 'Your nagios contact groups.'
		),
		'command' => array(
			'description' => 'Your nagios commands.'
		),
		'timeperiod' => array(
			'description' => 'Your nagios timeperiods.'
		),
		'downtime' => array(
			'description' => 'The hosts and services in scheduled downtime.',
			'nameless' => true
		),
		'comment' => array(
			'description' => 'Your nagios comments.',
			'nameless' => true
		),
		'status' => array(
			'description' => 'Nagios meta-information.',
			'nameless' => true
		)
	);

	/**
	 * @param $auth Op5Auth
	 */
	public function __construct(Op5Auth $auth) {
		$this->auth = $auth;
	}

	/**
	 * @throws Api_Error_Response
	 * @return filehandle, Livestatus socket
	 */
	private function open_livestatus_socket()
	{
		$sock = @fsockopen('unix:///opt/monitor/var/rw/live', null, $errno, $errstr);
		if ($errno) {
			$i = 0;
			while ($i++ < 5) {
				sleep(1);
				$sock = @fsockopen('unix:///opt/monitor/var/rw/live', null, $errno, $errstr);
				if (!$errno) {
					break;
				}
			}
		}
		if ($errno)
			throw new Api_Error_Response("Couldn't open livestatus socket", 500, $errstr);
		return $sock;
	}

	/**
	 * @return array
	 */
	private function get_menu()
	{
		$items = array();
		foreach($this->obj_types as $object_type => $properties) {
			$items[$object_type] = array(
				'resource' => "/api/help/status/$object_type"
			);
		}
		return array(
			array(
				'title' => 'Object types',
				'items' => $items
			)
		);
	}

	/**
	 * @param $socket filehandle Livestatus socket
	 * @param $len int
	 * @return string
	 */
	private function read_socket($socket, $len)
	{
		$offset = 0;
		$res = '';

		while ($offset < $len) {
			$data = fread($socket, $len - $offset);
			if (empty($data))
				break;
			$res .= $data;
			$offset += strlen($data);
		}
		return $res;
	}

	/**
	 * @param $request Api_Request
	 * @param $type string, Nagios object
	 * @return Api_Response
	 */
	private function html_response(Api_Request $request, $type)
	{
		$title = $type ? $type : 'op5 Monitor API > Help > Status';
		$body = null;
		$sock = $this->open_livestatus_socket();
		if(substr($type, -1) == 's') {
			// note the missing plural
			$query = "GET columns\nColumns: name type description\nFilter: table = {$type}\nOutputFormat: json\nResponseHeader: fixed16\n\n";
		} else {
			$query = "GET columns\nColumns: name type description\nFilter: table = {$type}s\nOutputFormat: json\nResponseHeader: fixed16\n\n";
		}
		@fwrite($sock, $query);
		$head = $this->read_socket($sock, 16);
		if (empty($head)) {
			fclose($sock);
			op5log::instance('http_api')->log('error', 'Invalid query: '.$query);
			return new Api_Error_Response("No response received from livestatus", 500);
		}
		$out = $this->read_socket($sock, substr($head, 4, 15));
		fclose($sock);
		if (substr($head, 0, 3) !== '200') {
			op5log::instance('http_api')->log('error', 'Invalid query: '.$query);
			return new Api_Error_Response("Internal error: ".$head, 500, $out);
		} else if (empty($out)) {
			op5log::instance('http_api')->log('error', 'Invalid query: '.$query);
			return new Api_Error_Response("No response received from livestatus", 500);
		}

		$ls_data = json_decode($out);
		$body .= "<table>
			<tr>
			<th>Property</th>
			<th>Type</th>
			<th>Description</th>
			</tr>";
		array_shift($ls_data);
		foreach ($ls_data as $row) {
			$body .= '<tr><td><var>'.$row[0].'</var></td><td>'.$row[1].'</td><td>'.$row[2].'</td></tr>';
		}
		$body .= '</table>';


		$example_xml_wget_uri = 'http'.(isset($_SERVER['HTTPS']) && 'on' == $_SERVER['HTTPS'] ? 's' : null).'://'.$_SERVER['HTTP_HOST']."/api/status/$type";
		$spelled_out_type_plural = rtrim($type, 's').'s';
		$is_real_nagios_object = $type != 'status';
		$example_part_of_object_name = array(
			'linux',
			'monitor',
			'norad',
			'joshua',
			'hal9000',
			'mother',
			'red-queen',
			'skynet'
		);
		$example_part_of_object_name = $example_part_of_object_name[array_rand($example_part_of_object_name)];
		$body .= "
			<h2>Live data</h2>
			<p>List real $type data from your Monitor installation in <a href='./../../status/$type?format=xml'>XML</a>, <a href='./../../status/$type?format=json'>JSON</a> or <a href='./../../status/$type?format=html'>HTML</a>.</p>";

		if($is_real_nagios_object) {
			$body .= "<p>If you would like to view a single object and its properties, edit the URI to point at <em>/api/status/$type/&lt;name of object here&gt;?format=xml</em>.</p>";
		}
		$body .= "

			<h2>Example usage</h2>";

		if($is_real_nagios_object) {
			$body .= "
				<p>Follow the examples below to extract $type objects remotely.</p>
				<h3>Limiting results</h3>
				<p>Include <em>?name=$example_part_of_object_name</em> in the URI to search for objects whose name contains <em>$example_part_of_object_name</em>. This works for all object types, even $spelled_out_type_plural.</p>";
		}
		$body .= "

			<h3 id='example_curl'>cURL in terminal</h3>
			<div class='curl'>
			<h4>GET</h4>
			<pre><code>curl '$example_xml_wget_uri?format=xml' -u '&lt;your username&gt;:&lt;your password&gt;'</code></pre>
			</div>

			<h3 id='example_python'>Python</h3>
			<div class='python'>
			<h4>GET</h4>
			<pre><code>import requests\nr = requests.get('$example_xml_wget_uri?format=json', auth=('&lt;your username&gt;', '&lt;your password&gt;'))\nprint r.text</code></pre>
			</div>

			<h3 id='example_php'>PHP with cURL</h3>
			<div class='php'>
			<h4>GET</h4>
			<pre><code>\$a_handle = curl_init('$example_xml_wget_uri?format=json');\ncurl_setopt(\$a_handle, CURLOPT_USERPWD, '&lt;your username&gt;:&lt;your password&gt;');\ncurl_setopt(\$a_handle, CURLOPT_RETURNTRANSFER, TRUE);\n\${$type}s = curl_exec(\$a_handle);\necho \${$type}s;</code></pre></div>";
		return new Api_Response(
			array(
				'body' => $body,
				'menu' => $this->get_menu()
			)
		);
	}

	/**
	 * @param $request Api_Request
	 * @return Api_Response
	 */
	private function index_html_response(Api_Request $request)
	{
		$title = 'op5 Monitor API > Help > Status';
		$body = "<h2>Using the <a href='http://en.wikipedia.org/wiki/Representational_state_transfer#RESTful_web_services' title='Representational state transfer'>REST</a> API</h2>

			<p class=meta>Since Monitor 6.3, you should use <a href='/api/help/filter'>/api/filter</a> instead of /api/status. It contains every feature that /api/status had, and more. For example, it lets you query the API for a list of objects with much more granularity. If you are about to code a new integration with the HTTP API, <strong>do use the <a href='/api/help/filter'>/api/filter</a></strong>.</p>

			<h3>Introduction</h3>
			<p>This is the REST-API that let's you query op5 Monitor for your objects health by issuing regular HTTP requests. Basically, you 'visit' an URI, which triggers Monitor to do something, and you get a response telling you what happened.</p>
			<p>You get help with the formulation of the requests on each object type's help page (look over to your left to find which object types are available). <strong>You also get tips on how to perform HTTP requests via the terminal (using cURL) and via scripting languages (using Python or PHP).</strong></p>

			<h3>Technical information</h3>

			<h4>Authentication</h4>
			<p>You send along a pair of regular op5 Monitor user credentials in a basic auth header with every request. View any example and you will understand how to do this.</p>

			<h4>Requests</h4>
			<p>In a REST-API, there are a couple of important things that make up for a valid request:</p>
			<ul>
			<li>The URI, in the following form
			<ul>
			<li>/api/status/host</li>
			<li>/api/status/host/a-hostname</li>
			</ul>
			</li>
			<li>The HTTP method
			<ul>
			<li>GET: lists objects and their current status.</li>
			</ul>
			</li>
			<li>The format as a HTTP header:
			<ul>
			<li><var>accept: application/xml</var></li>
			<li><var>accept: application/json</var> (default, except for requests from a browser)</li>
			<li><var>accept: text/html</var> (default in a browser environment)</li>
			</ul>
			or, appended to the URI:
			<ul>
			<li>?format=xml</li>
			<li>?format=json (default, except for requests from a browser)</li>
			<li>?format=html (default in a browser environment)</li>
			</ul>
			</li>
			<li>
			Filters. Restrict the list of objects (e.g. /api/status/host) by adding one ore more of the following parameters to the URI:
			<ul>
			<li>?state=1<br />Only include objects that are in Warning state</li>
			<li>?state[]=1&state[]=2<br />Only include objects that are in Warning or Critical state</li>
			<li>?state=1&hard=0<br />Only include objects that are in Warning, soft state</li>
			<li>?name=monitor<br />Only include objects whose name contains 'monitor'</li>
			<li>?name=linux&state=0&hard=1&format=xml<br />Only include objects that are in OK, hard state, whose name contains 'linux', get output in xml</li>
			<li>?hostgroup=servers<br />Only include objects that are attached to objects in hostgroup 'servers' (this only works if you are asking for services or hosts)</li>
			</ul>
			</li>
			</ul>

			<p>Once again, viewing any of the examples helps you get into the REST mindset.</p>

			<h4>Responses</h4>
			<p>You need to examine two things in the response: the status code and the body. The status code tells you what Monitor did with your input and the body gives you a detailed response.</p>

			<p>Examples of status codes:</p>
			<ul>
			<li>200: Everything went fine.</li>
			<li>400: You sent a request Monitor didn't understand. Correct the input and try again.</li>
			<li>401: You're not authenticated.</li>
			<li>403: You're not authorized to do this.</li>
			<li>404: The URI doesn't exist.</li>
			</ul>

			<p>These status codes are general for all web traffic and can be found at the <a href='http://en.wikipedia.org/wiki/HTTP_status_code'>List of HTTP status codes</a></p>";
		return new Api_Response(
			array(
				'body' => $body,
				'menu' => $this->get_menu()
			),
			200
		);
	}

	/**
	 * Nameless types differ from regular ones in that they don't have an
	 * object name. This means you can only list all objects in the system.
	 */
	private function handle_nameless(Api_Request $request, $type) {
		$sock = $this->open_livestatus_socket();
		$query = "GET ".rtrim($type, 's')."s\n";
		$username = $request->get_username();

		$mayi_resource = ObjectPool_Model::pool(rtrim($type, 's').'s')->all()->mayi_resource();
		mayi_run_api($mayi_resource.':read.%s');
		/* If not authorized for all_*, let livestatus find out if authorized anyway */
		if (!$this->auth->authorized_for('system_information'))
			$query .= "AuthUser: $username\n";
		$query .= "OutputFormat: json\nResponseHeader: fixed16\n\n";
		@fwrite($sock, $query);
		$head = $this->read_socket($sock, 16);
		if (empty($head)) {
			fclose($sock);
			op5log::instance('http_api')->log('error', 'Invalid query: '.$query);
			return new Api_Error_Response("No response received from livestatus", 500);
		}
		$out = $this->read_socket($sock, substr($head, 4, 15));
		fclose($sock);
		if (substr($head, 0, 3) !== '200') {
			op5log::instance('http_api')->log('error', 'Invalid query: '.$query);
			return new Api_Error_Response("Invalid request ".$head, 500, $out);
		} else if (empty($out)) {
			op5log::instance('http_api')->log('error', 'Invalid query: '.$query);
			return new Api_Error_Response("No response received from livestatus", 500);
		}

		$ls_data = json_decode($out);

		$response = array();
		$columns = array_shift($ls_data);
		if ($ls_data) {
			foreach ($ls_data as $params) {
				$row = array();
				foreach($params as $row_id => $param) {
					$row[$columns[$row_id]] = $param;
				}
				$response[] = $row;
			}
		}
		if($type == 'status') {
			// we only got a single row here
			$response = reset($response);
		}
		return new Api_Response($response);
	}

	/**
	 * @param $request Api_Request
	 * @param $type string nagios object type
	 * @param $name = false string name of object
	 * @return string|Api_Error_Response
	 */
	static function build_livestatus_query(Api_Request $request, $type, $name = false) {
		if ($name) {
			if ($type === 'service') {
				if(false === strpos($name, ';')) {
					return new Api_Error_Response('Wrong format of service name', 400);
				}
				list($name, $descr) = explode(';', $name);
				return "GET {$type}s\n" .
					"Filter: host_name = $name\n" .
					"Filter: description = $descr\n";
			}
			return "GET {$type}s\n" .
				"Filter: name = $name\n";
		}
		$params = $request->get_params();
		if ($type === 'service') {
			$query = "GET {$type}s\n" .
				"Columns: host_name description\n";
		} else {
			$query = "GET {$type}s\n" .
				"Columns: name\n";
		}
		if (isset($params['state'])) {
			if (is_array($params['state'])) {
				foreach ($params['state'] as $state) {
					$state = (int)$state;
					$query .= "Filter: state = $state\n";
				}
				$query .= "Or: ".count($params['state'])."\n";
			}
			else {
				$state = (int)$params['state'];
				$query .= "Filter: state = $state\n";
			}
		}
		if(isset($params['hostgroup'])) {
			$hostgroup = $params['hostgroup'];
			if (is_array($hostgroup)) {
				foreach ($hostgroup as $hg) {
					switch($type) {
					case 'service':
						$query .= "Filter: host_groups >= $hg\n";
						break;
					case 'host':
						$query .= "Filter: groups >= $hg\n";
						break;
					}
				}
				$query .= "Or: ".count($hostgroup)."\n";
			}
			else {
				switch($type) {
				case 'service':
					$query .= "Filter: host_groups >= $hostgroup\n";
					break;
				case 'host':
					$query .= "Filter: groups >= $hostgroup\n";
					break;
				}
			}
		}
		if (isset($params['hard'])) {
			$hard = (int)$params['hard'];
			$query .= "Filter: state_type = $hard\n";
		}
		if (isset($params['name'])) {
			if ($type === 'service') {
				$query .= "Filter: host_name ~ {$params['name']}\n";
				$query .= "Filter: description ~ {$params['name']}\n";
				$query .= "Or: 2\n";
			} else {
				$query .= "Filter: name ~ {$params['name']}\n";
			}
		}
		return $query;
	}

	/**
	 * @param $request Api_Request
	 * @param $type string, Nagios object type
	 * @param $name string = false, name of Nagios object
	 * @return Api_Response
	 * @throws Api_Error_Response for insufficient permissions
	 */
	private function handle_regular(Api_Request $request, $type, $name = false)
	{
		$query = self::build_livestatus_query($request, $type, $name);
		if($query instanceof Api_Error_Response) {
			return $query;
		}
		# let's say system_information for unrestricted browsing of timeperiods, command and contact stuffs
		switch($type) {
		case 'host':
		case 'hostgroup':
		case 'service':
		case 'servicegroup':
			$auth_type = $type . '_view_all';
			break;
		default:
			$auth_type = 'system_information';
			break;
		}

		$mayi_resource = ObjectPool_Model::pool(rtrim($type, 's').'s')->all()->mayi_resource();
		mayi_run_api($mayi_resource.':read.%s');

		/* If not authorized for all_*, let livestatus find out if authorized anyway (through contacts) */
		if (!$this->auth->authorized_for($auth_type)) {
			$username = $request->get_username();
			$query .= "AuthUser: $username\n";
		}

		$query .= "OutputFormat: json\nResponseHeader: fixed16\n\n";
		$sock = $this->open_livestatus_socket();
		@fwrite($sock, $query);
		$head = $this->read_socket($sock, 16);
		if (empty($head)) {
			fclose($sock);
			return new Api_Error_Response("No response received from livestatus", 500);
		}
		$out = $this->read_socket($sock, substr($head, 4, 15));
		fclose($sock);
		if (substr($head, 0, 3) !== '200')
			return new Api_Error_Response("Invalid request ".$head, 500, $out);
		else if (empty($out))
			return new Api_Error_Response("No response received from livestatus", 500);

		$ls_data = json_decode($out);

		if ($name) {
			if (count($ls_data) == 1) {
				return new Api_Error_Response("Not Found", 404);
			}

			$response = array();
			foreach ($ls_data[0] as $idx => $param) {
				$response[$param] = $ls_data[1][$idx];
				if (isset($this->obj_types[$type]['duplicate'][$param])) {
					$response[$this->obj_types[$type]['duplicate'][$param]] = $ls_data[1][$idx];
				}
			}
		}
		else {
			$response = array();
			$base_url = $request->get_base_url().'/api/status/';
			if ($type === 'service') {
				foreach ($ls_data as $obj) {
					$response[] = array('name' => $obj[0].';'.$obj[1], 'resource' => $base_url.'service/'.urlencode(urlencode($obj[0].';'.$obj[1])));
				}
			} else {
				foreach ($ls_data as $obj) {
					$response[] = array('name' => $obj[0], 'resource' => $base_url.$type.'/'.urlencode(urlencode($obj[0])));
				}
			}
		}
		return new Api_Response($response);
	}

	/**
	 * Main entry point for requests that should be handled by livestatus
	 *
	 * @param $request The HTTP request we're responding to
	 * @returns Api_Response
	 * @throws Api_Response
	 */
	public function handle_request(Api_Request $request)
	{
		$path = $request->get_path();
		$type = false;
		if (!empty($path))
			$type = array_shift($path);

		$name = false;
		if (!empty($path))
			$name = array_shift($path);
		if (!empty($path))
			return Api_Error_Response::with_usage('Invalid request', 404, $request);

		if ($type && array_key_exists($type, $this->obj_types)) {
			if($request->is_help()) {
				return $this->html_response($request, $type);
			}
			if ($_SERVER['REQUEST_METHOD'] !== 'GET')
				return new Api_Error_Response('Not implemented', 501);
			if (isset($this->obj_types[$type]['nameless'])) {
				if ($name)
					return Api_Error_Response::with_usage('Invalid request', 404, $request);
				return $this->handle_nameless($request, $type);
			}
			return $this->handle_regular($request, $type, $name);
		} elseif(!$type && $request->is_help()) {
			return $this->index_html_response($request);
		}
		return Api_Error_Response::with_usage('Invalid request', 404, $request);
	}
}
