<?php

/**
 * @param $request Api_Request
 * @param $auth op5auth
 * @return Api_Response
 */
function generate_response(Api_Request $request, op5auth $auth) {
	$log = op5log::instance('http_api');
	$api = new Perfdata_Api($auth, $log);
	return $api->handle_request($request);
}

/**
 * Controller for the HTTP API to manage /api/perfdata
 */
class Perfdata_Api
{
	const NAN = 'NaN'; /**< perfdata's "NaN" differs based on system, we should be consistent */
	const DEFAULT_CF = 'AVERAGE';
	private $auth;
	public $base_url; /**< The URI which all printed "/api/yadda yadda" links are based on */
	public $rrdcached_socket = 'unix:/opt/monitor/var/rrdtool/rrdcached/rrdcached.sock';
	private $log;
	private $xml_base_dir = '/opt/monitor/op5/pnp/perfdata/';
	protected $troubleshooting = array();

	/**
	 * @param $auth Op5Auth
	 */
	function __construct(Op5Auth $auth, op5LogAccess $log) {
		$this->auth = $auth;
		$this->log = $log;
	}

	/**
	 * @param $cmd array For example array("ls", "-l"). Easier to escape input if it's paired like this
	 * @param &$stdout str
	 * @param &$stderr str
	 * @param &$exit_code int
	 * @throws Api_Error_Response on 'deep' errors, instead of hijacking stderr for non-stderr related error messages
	 * @return void
	 */
	private function _exec(array $cmd, &$stdout, &$stderr, &$exit_code) {
		$stdout = $stderr = $exit_code = null; // reset variables by reference to avoid unexpected state
		$descriptorspec = array (
			0 => array("pipe","r"),
			1 => array("pipe","w"),
			2 => array("pipe","w")
		);

		$escaped_cmd = implode(" ", array_map('escapeshellarg', $cmd));
		$this->log->log('debug', sprintf('perfdata cmd: %s', $escaped_cmd));
		$process = proc_open($escaped_cmd, $descriptorspec, $pipes);
		if (!is_resource($process)) {
			$this->log->log('warning', sprintf('Could not start process (should have been detected before?): %s', $escaped_cmd));
			throw new Api_Error_Response("Internal server error", 500, "Try again, contact op5 Support if the problem persists (proc open fail)");
		}
		fwrite($pipes[0], $escaped_cmd);
		fclose($pipes[0]);
		stream_set_timeout($pipes[1], 1);
		$stdout = trim(stream_get_contents($pipes[1]));
		stream_set_timeout($pipes[2], 1);
		$stderr = stream_get_contents($pipes[2]);
		$stream_meta = stream_get_meta_data($pipes[1]);
		if($stream_meta['timed_out']){
			$this->log->log('error', sprintf('Timeout (1 second): %s', $escaped_cmd));
			throw new Api_Error_Response("Internal server error", 500, "Try again, contact op5 Support if the problem persists (popen timeout)");
		}
		fclose($pipes[1]);
		fclose($pipes[2]);
		$exit_code = proc_close($process);
		if($exit_code) {
			$this->log->log('debug', sprintf('Exit code %d, stdout: %s, stderr: %s', $exit_code, $stdout, $stderr));
		}
	}


	/**
	 * Used in a help context
	 *
	 * @return array
	 */
	function get_menu()
	{
		return array(
			'title' => _('Sections'),
			'items' => array_map(function($item) {
					$item['resource'] = preg_replace("~/api/~", "/api/help/", $item['resource']);
					return $item;
				},
				$this->sections()
			)
		);
	}

	/**
	 * @param $xml_filename string
	 * @return array [ ds1, ds2, ..., dsN ]
	 * @throws Api_Error_Response
	 */
	function get_ds_for_service($xml_filename) {
		$xml = simplexml_load_file($xml_filename);
		if(!$xml) {
			$this->log->log('error', "Could not parse XML in $xml_file");
			throw new Api_Error_Response("Internal server error", 500, "Try again, contact op5 Support if the problem persists (the XML could not be parsed)");
		}
		$datasources = array();
		foreach($xml->DATASOURCE as $ds) {
			$datasources[] = (string) $ds->NAME;
		}
		// TODO this is where we can add more properties that's good to
		// know about a ds, such as 'warn_min' et al
		return $datasources;
	}

	/**
	 * @param $request Api_Request
	 * @return Api_Response
	 */
	function handle_fetch(Api_Request $request) {
		// cf first, because it does not acces the file system
		$cf = $request->get_param('cf', self::DEFAULT_CF);
		$cfs = array('AVERAGE', 'MIN', 'MAX', 'LAST');
		if(!in_array($cf, $cfs, true)) {
			return new Api_Error_Response('Invalid consolidation function', 400, "Must be one of ".implode(", ", $cfs));
		}

		// host
		$host = $request->get_param('host', false);
		if($host === false || $host == '.pnp-internal') {
			return new Api_Error_Response('No host given', 400);
		}
		$host_set = HostPool_Model::all()->reduce_by('name', $host, '=');
		if($host_set->count() != 1) {
			return new Api_Error_Response('Host does not exist', 404, "Did you urlencode all parameters? Are you authenticated to view this object?");
		}
		if(!is_dir($this->xml_base_dir.pnp::clean($host))) {
			return new Api_Error_Response('No perfdata found for host', 404, "It seems the host you asked for, $host, has not yet generated any perfdata.");
		}

		// service
		$service = $request->get_param('service', false);
		if($service === false) {
			$service_set = ServicePool_Model::all()->reduce_by('host.name', $host, '=');
			$service_data = array();
			foreach($service_set->it(array('description')) as $service) {
				$service_data[] = array(
					'name' => $host.";".$service->get_description(),
					'resource' => $request->get_base_url().'/api/beta/perfdata/fetch/?host='.rawurlencode($host).'&service='.rawurlencode($service->get_description())
				);
			}
			return new Api_Response($service_data);
		}
		$set = ServicePool_Model::all()->reduce_by('host.name', $host, '=')->reduce_by('description', $service, '=');
		if($set->count() != 1) {
			return new Api_Error_Response('Service does not exist', 404, "Did you urlencode all parameters? Are you authenticated to view this object?");
		}
		$service_xml_file = $this->xml_base_dir.pnp::clean($host).'/'.pnp::clean($service).'.xml';
		if(!is_readable($service_xml_file) || realpath($service_xml_file) != $service_xml_file) {
			return new Api_Error_Response('No perfdata found for host', 404, "It seems the service you asked for, $host;$service, has not yet generated any perfdata.");
		}

		// ds
		$ds = $request->get_param('ds', false);
		if($ds === false) {
			$ds = $this->get_ds_for_service($service_xml_file);
			$ds = array_map(function($datasource) use ($host, $service, $request) {
				return array(
					'name' => $datasource,
					'resource' => $request->get_base_url().'/api/beta/perfdata/fetch/?host='.rawurlencode($host).'&service='.rawurlencode($service).'&ds='.rawurlencode($datasource)
				);
			}, $ds);
			return new Api_Response($ds);
		}
		$rrd_file_path = '/opt/monitor/op5/pnp/perfdata/'.pnp::clean($host).'/'.pnp::clean($service).'_'.$ds.'.rrd';
		if(!is_readable($rrd_file_path) || $rrd_file_path != realpath($rrd_file_path)) {
			return new Api_Error_Response('DS does not exist', 404, "Did you urlencode all parameters?");
		}

		$cmd = array(
			'rrdtool',
			'fetch',
			'--daemon',
			$this->rrdcached_socket,
			$rrd_file_path,
			$cf
		);
		if(false !== ($start = $request->get_param('start', false))) {
			if(!preg_match('/\d{10}/', $start)) {
				$start = strtotime($start);
				if($start === false) {
					return new Api_Error_Response('Invalid value for start', 400, "It must be a UNIX timestamp or a string that can be parsed with the PHP function strtotime(). See http://php.net/manual/en/datetime.formats.php for a list of valid formats");
				}
			}
			$cmd[] = '--start';
			$cmd[] = $start;
		}
		if(false !== ($end = $request->get_param('end', false))) {
			if(!preg_match('/\d{10}/', $end)) {
				$end = strtotime($end);
				if($end === false) {
					return new Api_Error_Response('Invalid value for end', 400, "It must be a UNIX timestamp or a string that can be parsed with the PHP function strtotime(). See http://php.net/manual/en/datetime.formats.php for a list of valid formats");
				}
			}
			$cmd[] = '--end';
			$cmd[] = $end;
		}
		if($start !== false && $end !== false && $start >= $end) {
			return new Api_Error_Response('Invalid timerange', 400, "Start must occur earlier than the end");
		}
		$this->_exec($cmd, $stdout, $stderr, $exit_code);
		if($exit_code)
			return new Api_Error_Response("Internal server error", 500, "Try again, contact op5 Support if the problem persists (process exit code)");

		// rrdtool output begins with a header: "\s+ds_name_here\n\n"
		$output = array_splice(preg_split('/\n/', $stdout), 2);
		$real_output = array();
		foreach($output as $i => $row) {
			// "int: int" (where the second int is in scientific notation)
			$split = explode(': ', $row);
			$value = $split[1];
			if(false !== stripos($value, 'nan')) {
				// normalize system dependent styling of "NaN"
				// to our definition
				$value = self::NAN;
			}
			$real_output[$split[0]] = $value;
		}
		return new Api_Response($real_output);
	}

	/**
	 * @param $section string
	 * @param $request Api_Request
	 * @return Api_Response
	 */
	function help($section, Api_Request $request)
	{
		$sections = $this->sections();
		$response = array(
			'menu' => $this->get_menu()
		);
		switch($section) {
			case 'index':
				$response['body'] = file_get_contents(__DIR__.'/perfdata/index.html');
				break;
			case 'fetch':
				$response['body'] = file_get_contents(__DIR__.'/perfdata/fetch.html');
				$response['properties'] =  array(
					array(
						'name' => 'host',
						'type' => 'string',
						'default' => '',
						'description' => 'The name of a host',
						'required' => true
					),
					array(
						'name' => 'service',
						'type' => 'string',
						'default' => '',
						'description' => 'The name of a service',
						'required' => true
					),
					array(
						'name' => 'ds',
						'type' => 'string',
						'default' => '',
						'description' => 'The name of a datasource that exists on the wanted host or service',
						'required' => true
					),
					array(
						'name' => 'cf',
						'type' => 'enum',
						'default' => 'AVERAGE',
						'description' => 'Consolidation function, one of AVERAGE, MIN, MAX, LAST',
						'required' => false
					),
					array(
						'name' => 'start',
						'type' => 'mixed',
						'default' => 'One day ago',
						'description' => 'A value for the earliest data you are interested in. Can be a UNIX timestamp but also <a target=_blank href="http://php.net/manual/en/datetime.formats.php">anything that PHP\'s strtotime() can parse.</a>',
						'required' => false
					),
					array(
						'name' => 'end',
						'type' => 'mixed',
						'default' => 'now',
						'description' => 'A value for the earliest data you are interested in. Can be a UNIX timestamp but also <a target=_blank href="http://php.net/manual/en/datetime.formats.php">anything that PHP\'s strtotime() can parse.</a>',
						'required' => false
					),
				);
				break;
			default:
				$response['body'] = $sections[$section]['description'];
		}
		$response['troubleshooting'] = $this->troubleshooting;
		return new Api_Response($response);
	}

	/**
	 * Main entry point for requests that should be handled by livestatus
	 *
	 * @param $request The HTTP request we're responding to
	 * @returns Api_Response
	 * @throws Api_Response
	 */
	function handle_request(Api_Request $request)
	{
		$path = $request->get_path();
		$this->base_url = $request->get_base_url().'/api/beta/perfdata/';
		$sections = $this->sections();
		if(!$path) {
			return $this->help('index', $request);
		}
		$section = array_shift($path);
		if($section && isset($sections[$section]) && $request->is_help()) {
			return $this->help($section, $request);
		}
		switch($section) {
			case 'fetch':
				return $this->{'handle_'.$section}($request);
			default:
				return Api_Error_Response::with_usage("Invalid request, missing subsection like 'fetch'", 400, $request);
		}
	}

	/**
	 * @return array
	 */
	function sections()
	{
		return array(
			'fetch' => array(
				'description' => file_get_contents(__DIR__.'/perfdata/fetch.html'),
				'resource' => $this->base_url.'fetch'
			),
		);
	}
}
