<?php

/**
 * @param $request Api_Request
 * @param $auth op5auth
 * @return Api_Response
 */
function generate_response(Api_Request $request, op5auth $auth) {
	$api = new Report_Api($auth);
	return $api->handle_request($request);
}

/**
 * Report section of the HTML API, façade/adapter for Ninja's report layer
 * over Merlin. Takes requests as input, leaves responses.
 */
class Report_Api
{
	private $auth;
	public $base_url; /**< The URI which all printed "/api/yadda yadda" links are based on */

	/**
	 * Array of hints to include in the web docs
	 */
	protected $troubleshooting = array(
		"slow_acl" => array(
			"title" => "How can I make the report API faster? Is there a good workflow for working with the API?",
			"content" => "Having to authorize the calling user for each host/service generates some load on the server. You can avoid the authorization performance penalty by giving your user the rights 'host_view_all' and 'service_view_all'. Note that those settings are very allowing so be careful when you're giving your users those rights. To separate concerns, consider using a dedicated report API user that you can lock down and rely on being used for this special purpose only."
		)
	);

	/**
	 * @param $auth Op5Auth
	 */
	function __construct(Op5Auth $auth) {
		$this->auth = $auth;
	}

	/**
	 * Used in a help context
	 *
	 * @return array
	 */
	function get_menu()
	{
		return array(
			'title' => _('Sections'),
			'items' => array_map(function($item) {
					$item['resource'] = preg_replace("~/api/~", "/api/help/", $item['resource']);
					return $item;
				},
				$this->sections()
			)
		);
	}

	/**
	 * Handle unfiltered get, and return an array of initial state representations.
	 *
	 * @param $options array
	 * @return array
	 */
	function handle_state($options)
	{
		if(!$options) {
			throw new Api_Error_Response("No options provided", 400);
		}
		if (!$options['time']) {
			throw new Api_Error_Response("No time provided to show state at", 400);
		}
		try {
			$options_obj = Report_options::setup_options_obj('HttpApiState', $options);
		} catch (ReportValidationException $ex) {
			throw new Api_Error_Response($ex->getMessage(), 400);
		}
		if (!$options_obj['objects']) {
			throw new Api_Error_Response("No object provided to show state for", 400);
		}
		$reports = new Status_Reports_Model($options_obj);
		$result = $reports->get_state();

		$output_options = $options;

		$data = array();
		foreach($result as $row) {
			$data[] = $options_obj->to_output($row);
		}

		ksort($output_options); // easier to remember parameters if their order is constant
		$response = array(
			'resource' => $this->base_url."event/?".http_build_query($output_options),
			'generated_at' => time(),
			'time' => arr::search($options, 'time', null),
		);
		// make sure all options are prepended, and the data appended; keep the meta in the 'head'
		$response['data'] = $data;

		return $response;
	}

	/**
	 * "options" as in "unfiltered GET" -- let's filter those arguments here and pass them
	 * along to Summary_Reports_Model::get_events()
	 *
	 * @param $options array
	 * @return array
	 */
	function handle_event($options)
	{
		if(!$options) {
			throw new Api_Error_Response("No options provided", 400);
		}
		if(isset($options['report_period']) && $options['report_period'] != 'custom' && (isset($options['start_time']) || isset($options['end_time']))) {
			throw new Api_Error_Response(sprintf("Can not combine 'report_period' (%s) with 'start_time' or 'end_time'", $options['report_period']), 400);
		}
		try {
			$options_obj = Report_options::setup_options_obj('HttpApiEvent', $options);
		} catch (ReportValidationException $ex) {
			throw new Api_Error_Response($ex->getMessage(), 400);
		}
		$reports = new Summary_Reports_Model($options_obj);
		$result = $reports->get_events($this->auth);
		$data = array();
		foreach($result['events'] as $row) {
			$data[] = $options_obj->to_output($row);
		}

		if ($options_obj['start_time'])
			$options['start_time'] = (string)$options_obj['start_time'];
		if ($options_obj['end_time'])
			$options['end_time'] = (string)$options_obj['end_time'];
		if ($options_obj['include_downtime'])
			$options['include_downtime'] = (int)$options_obj['include_downtime'];

		// let's not confuse one with the other
		$output_options = $options;
		unset($output_options['report_period']);

		ksort($output_options); // easier to remember parameters if their order is constant
		$response = array(
			'resource' => $this->base_url."event/?".http_build_query($output_options),
			'generated_at' => time(),
			'start_time' => arr::search($output_options, 'start_time', null),
			'end_time' => arr::search($output_options, 'end_time', null),
			'start_id' => arr::search($output_options, 'start_id', null),
			'end_id' => arr::search($output_options, 'end_id', null),
			'limit' => $result['limit'],
			'offset' => $result['offset']
		);
		if($result['can_paginate']) {
			// provide easy pagination through additional resource pointer (aka "link")
			$output_options['offset'] = $result['offset'] + $result['limit'];
			ksort($output_options); // easier to remember parameters if their order is constant
			$response['next_resource'] = $this->base_url."event/?".http_build_query($output_options);
			// adhere to the limit given and do the cheap thing when appreciating
			// the need for pagination: remove the extra element
			unset($data[count($data)-1]);
		}
		// make sure all options are prepended, and the data appended; keep the meta in the 'head'
		$response['data'] = $data;
		return $response;
	}

	/**
	 * @param $section string
	 * @param $request Api_Request
	 * @return Api_Response
	 */
	function help($section, Api_Request $request)
	{
		$sections = $this->sections();
		$response = array(
			'menu' => $this->get_menu()
		);
		$format_type = function($type, $enums) {
			if(!$enums) {
				$enums = array("value1", "value2");
			}
			$map = array(
				'int' => 'A number',
				'string' => 'Free text',
				'enum' => "One of ['".implode("', '", $enums)."']",
				'bool' => "1 (meaning 'yes') or 0 (meaning 'no')",
				'objsel' => "One or more names of objects",
				'array' => "One or more of ['".implode("', '", $enums)."']",
				'timestamp' => "A Unix timestamp (it's now ".time()." for example)",
				'object' => ''
			);
			return sprintf('<abbr title="%s">%s</abbr>', $map[$type], $type == 'objsel' ? 'array' : $type);
		};
		switch($section) {
			case 'index':
				$response['body'] = file_get_contents(__DIR__.'/report/index.html');
				break;
			case 'state':
				try {
					$options = Report_options::setup_options_obj('HttpApiState', $request->get_params());
				} catch (ReportValidationException $ex) {
					throw new Api_Error_Response($ex->getMessage(), 400);
				}
				// @todo put all of this in HttpApiEvent_options
				$properties = $options->properties();
				$required_names = array(
					'time',
				);
				$rows = array();
				foreach($properties as $name => $option) {
					if (isset($option['generated']))
						continue;
					$key = "required_$name";
					$required = true;
					if(!in_array($name, $required_names)) {
						$key = 'zzzz'.$key; // place it later than the required keys
						$required = false;
					}
					$description = null;
					if(isset($option['description'])) {
						$description = $option['description'];
					}

					$options_to_print = $options->get_alternatives($name);
					$rows[$key] = array(
						'required' => $required,
						'name' => $name,
						'type' => $format_type($option['type'], $options_to_print),
						'default' => $options->format_default($name, $option['type'] == 'objsel' ? 'array' : $option['type']),
						'description' => $description
					);
				}

				ksort($rows);

				$host = HostPool_Model::all()->it(array('name'))->current()->get_name();

				$live_data = sprintf(
					'<h2>Live data</h2>
					<p>List real objects from your Monitor installation in <a href="%1$sstate/?format=xml&time=%2$s&host_name[]=%3$s">XML</a>, <a href="%1$sstate/?format=json&time=%2$s&host_name[]=%3$s">JSON</a> or <a href="%1$sstate/?format=html&time=%2$s&host_name[]=%3$s">HTML</a>.</p>
					If you would like to view a single object and its properties, edit the URI to point at /api/report/?host_name=&lt;name of object here&gt;&format=xml',
					$this->base_url,
					time(),
					$host
				);

				$example_uri = $request->get_base_url().'/api/report/state';

				$example = json::pretty($this->handle_state(array('time' => time(), 'host_name' => array($host))));
				$example_usage = "
				<h2>Example usage</h2>
				<p>Follow the examples below to extract events.</p>

				<h3 id='example_curl'>cURL in terminal</h3>
				<div class='curl'>
					<h4>GET</h4>
					<h5>Request</h5>
					<pre><code>curl '$example_uri/?host_name[]=$host&amp;time='$(date +%s) -u '&lt;your username&gt;:&lt;your password&gt;'</code></pre>
					<h5>Response</h5>
					<pre><samp>".$example."</samp></pre>
				</div>
				<h3 id='example_python'>Python</h3>
				<div class='python'>
					<h4>GET</h4>
					<pre><code>import requests\nr = requests.get('$example_uri/?host_name[]=$host', auth=('&lt;your username&gt;', '&lt;your password&gt;'))\nprint r.text</code></pre>
				</div>

				<h3 id='example_php'>PHP with cURL</h3>
				<div class='php'>
					<h4>GET</h4>
					<pre><code>\$a_handle = curl_init('$example_uri/?host_name[]=$host');\ncurl_setopt(\$a_handle, CURLOPT_USERPWD, '&lt;your username&gt;:&lt;your password&gt;');\ncurl_setopt(\$a_handle, CURLOPT_RETURNTRANSFER, TRUE);\n\$events = curl_exec(\$a_handle);\necho \$events;</code></pre>
				</div>
					";

				$response['body'] = implode("\n", array(
					$live_data,
					$example_usage
				));
				$response['description'] = $sections[$section]['description'];
				$response['properties'] = $rows;
				break;
			case 'event':
				try {
					$options = Report_options::setup_options_obj('HttpApiEvent', $request->get_params());
				} catch (ReportValidationException $ex) {
					throw new Api_Error_Response($ex->getMessage(), 400);
				}
				// @todo put all of this in HttpApiEvent_options
				$properties = $options->properties();
				$required_names = array(
					'end_id',
					'start_id',
					'end_time',
					'start_time'
				);
				$rows = array();
				foreach($properties as $name => $option) {
					if (isset($option['generated']))
						continue;
					$key = "required_$name";
					$required = true;
					if(!in_array($name, $required_names)) {
						$key = 'zzzz'.$key; // place it later than the required keys
						$required = false;
					}
					$description = null;
					if(isset($option['description'])) {
						$description = $option['description'];
					}

					$options_to_print = $options->get_alternatives($name);
					$rows[$key] = array(
						'required' => $required,
						'name' => $name,
						'type' => $format_type($option['type'], $options_to_print),
						'default' => $options->format_default($name, $option['type'] == 'objsel' ? 'array' : $option['type']),
						'description' => $description
					);
				}

				ksort($rows);

				$id = Database::instance()->query('select id from report_data order by id desc limit 1 offset 10')->result()->current();
				if ($id)
					$id = $id->id;
				else
					$id = 1;
				$live_data = sprintf(
					'<h2>Live data</h2>
					<p>List real objects from your Monitor installation in <a href="%1$sevent/?format=xml&start_id=%2$s&limit=10">XML</a>, <a href="%1$sevent/?format=json&start_id=%2$s&limit=10">JSON</a> or <a href="%1$sevent/?format=html&start_id=%2$s&limit=10">HTML</a>.</p>
					If you would like to view a single object and its properties, edit the URI to point at /api/report/event?host_name=&lt;name of object here&gt;&format=xml',
					$this->base_url,
					$id
				);

				$example_uri = $request->get_base_url().'/api/report/event';

				$example = json::pretty($this->handle_event(array('start_id' => $id)));

				$local_limit = (int) Op5Config::instance()->getConfig('http_api.report.limit');
				if(!$local_limit) {
					$local_limit = "<em>not set</em>";
				}

				$python_pagination_example = file_get_contents(__DIR__.'/examples/traverse.py');
				$example_usage = "
				<h2>Example usage</h2>
				<p>Follow the examples below to extract events.</p>

				<h3 id='example_curl'>cURL in terminal</h3>
				<div class='curl'>
					<h4>GET</h4>
					<h5>Request</h5>
					<pre><code>curl '$example_uri/?start_id=$id&limit=2' -u '&lt;your username&gt;:&lt;your password&gt;'</code></pre>
					<h5>Response</h5>
					<pre><samp>".$example."</samp></pre>
				</div>
				<h3 id='example_python'>Python</h3>
				<div class='python'>
					<h4>GET</h4>
					<pre><code>import requests\nr = requests.get('$example_uri/?start_id=$id', auth=('&lt;your username&gt;', '&lt;your password&gt;'))\nprint r.text</code></pre>
				</div>

				<h3 id='example_php'>PHP with cURL</h3>
				<div class='php'>
					<h4>GET</h4>
					<pre><code>\$a_handle = curl_init('$example_uri/?start_id=$id);\ncurl_setopt(\$a_handle, CURLOPT_USERPWD, '&lt;your username&gt;:&lt;your password&gt;');\ncurl_setopt(\$a_handle, CURLOPT_RETURNTRANSFER, TRUE);\n\$events = curl_exec(\$a_handle);\necho \$events;</code></pre>
				</div>

				<h2>Limiting results</h2>

				<dl>
					<dt id='example_traverse'><var>?limit=10&offset=10</var></dt>
					<dd>
						<p>A hard limit which <em>might</em> be adhered to. The limit is detected in the following order:</p>
						<ol>
							<li>Absolute max limit: ".HttpApiEvent_options::MAX_EVENTS."</li>
							<li>Local op5 Monitor max limit: ".$local_limit." (can be modified in /etc/op5/http_api.yml)</li>
							<li>Your submitted <var>limit</var> GET parameter</li>
						</ol>
						<p>The complementing <var>offset</var> makes sure you can always get all of your data, by performing requests with a limit lower than the total count, and an increasing offset, until you get all of it. Here's an example python script which does this for you (you will find all examples in the folder ".__DIR__."/examples):</p>
						<div class='python'>
							<pre><code>$python_pagination_example</code></pre>
						</div>
					</dd>

					<dt><var>?report_period=last24hours</var></dt>
					<dd>
						Limit the report period to a fixed timeframe. This lets you take shortcuts if you're only interested in, say, events that occurred during the <var>last24hours</var>. All other parameters in the table above are there for you to filter events.
					</dd>
				</dl>
					";

				$response['body'] = implode("\n", array(
					$live_data,
					$example_usage
				));
				$response['description'] = $sections[$section]['description'];
				$response['properties'] = $rows;
				break;
			default:
				$response['body'] = $sections[$section]['description'];
		}
		$response['troubleshooting'] = $this->troubleshooting;
		return new Api_Response($response);
	}

	/**
	 * Main entry point for requests that should be handled by livestatus
	 *
	 * @param $request The HTTP request we're responding to
	 * @returns Api_Response
	 * @throws Api_Response
	 */
	function handle_request(Api_Request $request)
	{
		$path = $request->get_path();
		$this->base_url = $request->get_base_url().'/api/report/';
		$sections = $this->sections();
		if(!$path) {
			return $this->help('index', $request);
		}
		$section = array_shift($path);
		if($section && isset($sections[$section]) && $request->is_help()) {
			return $this->help($section, $request);
		}
		switch($section) {
			case 'event':
				return new Api_Response($this->handle_event($request->get_params()));
			case 'state':
				return new Api_Response($this->handle_state($request->get_params()));
			default:
				return Api_Error_Response::with_usage('Invalid request', 400, $request);
		}
	}

	/**
	 * @return array
	 */
	function sections()
	{
		return array(
			'event' => array(
				'description' => '
				<p>The report-event part of Monitor\'s API represents the <strong>extract</strong> part of <a href="https://en.wikipedia.org/wiki/Extract_transform_load" title="Extract, Transform, Load">ETL</a>:</p>
				<ol>
					<li class="highlight"><strong>Extract</strong><br />Take report data out of Monitor in XML or JSON. You get the same information that Monitor saves and displays in the GUI, e.g. what happened, when it happened, which service/host was involved.</li>
					<li><strong>Transform</strong><br />Adapt the Monitor-formatted data to fit your target destination, turning it into valid input for your external program, e.g. another report generator.</li>
					<li><strong>Load</strong><br />Use your external report program with Monitor\'s data.</li>
				</ol>
				<p><strong>Note</strong> that while this API, in combination with the <a href="'.preg_replace("~/api/~", "/api/help/", $this->base_url).'state">state</a> API, might be useful for <em>light</em> real-time use, heavy use of the real-time features - or use of the real-time features in large environments - might be detrimental to your monitoring experience. If you experience performance problems, you should turn to a dedicated data transformation tool to generate your reports.</p>

				',
				'resource' => $this->base_url.'event'
			),
			'state' => array(
				'description' => '
				<p>The report-state part of Monitor\'s API provides you with a description of your monitored objects\' state at one specific point in time. The response is the state at the <i>beginning</i> of the second you ask for, so events that happened during that second are <i>not</i> included.</p>
				<p>The idea is that it will allow you to prime calculations while making <em>light</em> real-time use of the <a href="'.preg_replace("~/api/~", "/api/help/", $this->base_url).'event">event</a> API. Heavy use could be detrimental to your monitoring experience. If you experience performance problems, you should turn to a dedicated data transformation tool to generate your reports.</p>
				',
				'resource' => $this->base_url.'state'
			)
		);
	}
}
